# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Mathieu Clabaut <mathieu.clabaut@gmail.com>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import email.utils
import htmlentitydefs
import itertools
import os
import os.path
import re
import string
import time
import xml.sax.saxutils

from jinja2 import Environment, FileSystemLoader, DictLoader, ChoiceLoader

import libbe
import libbe.command
import libbe.command.depend
import libbe.command.target
import libbe.command.util
import libbe.comment
import libbe.util.encoding
import libbe.util.id
import libbe.util.wsgi
import libbe.version


class ServerApp (libbe.util.wsgi.WSGI_AppObject,
                 libbe.util.wsgi.WSGI_DataObject):
    """WSGI server for a BE Storage instance over HTML.

    Serve browsable HTML for public consumption.  Currently everything
    is read-only.
    """
    server_version = 'BE-html-server/' + libbe.version.version()

    def __init__(self, bugdirs={}, template_dir=None, title='Site Title',
                 header='Header', index_file='', min_id_length=-1,
                 strip_email=False, generation_time=None, **kwargs):
        super(ServerApp, self).__init__(
            urls=[
                (r'^{0}$'.format(index_file), self.index),
                (r'^style.css$', self.style),
                (r'^([^/]+)/([^/]+)/{0}'.format(index_file), self.bug),
                ],
            **kwargs)
        self.bugdirs = bugdirs
        self.title = title
        self.header = header
        self._index_file = index_file
        self.min_id_length = min_id_length
        self.strip_email = strip_email
        self.generation_time = generation_time
        self._refresh = 0
        self._load_templates(template_dir=template_dir)
        self._filters = {
            'active': lambda bug: bug.active and bug.severity != 'target',
            'inactive': lambda bug: not bug.active and bug.severity !='target',
            'target': lambda bug: bug.severity == 'target'
            }

    # handlers
    def style(self, environ, start_response): 
        template = self.template.get_template('style.css')
        content = template.render()
        return self.ok_response(
            environ, start_response, content, content_type='text/css')

    def index(self, environ, start_response):
        data = self.query_data(environ)
        source = 'query'
        bug_type = self.data_get_string(
            data, 'type', default='active', source=source)
        assert bug_type in ['active', 'inactive', 'target'], bug_type
        self.refresh()
        filter_ = self._filters.get(bug_type, self._filters['active'])
        bugs = list(itertools.chain(*list(
                    [bug for bug in bugdir if filter_(bug)]
                    for bugdir in self.bugdirs.values())))
        bugs.sort()
        if self.logger:
            self.logger.log(
                self.log_level, 'generate {0} index file for {1} bugs'.format(
                    bug_type, len(bugs)))
        template_info = {
            'title': self.title,
            'charset': 'UTF-8',
            'stylesheet': 'style.css',
            'header': self.header,
            'active_class': 'tab nsel',
            'inactive_class': 'tab nsel',
            'target_class': 'tab nsel',
            'bugs': bugs,
            'bug_entry': self.template.get_template('index_bug_entry.html'),
            'bug_dir': self.bug_dir,
            'index_file': self._index_file,
            'generation_time': self._generation_time(),
            }
        template_info['{0}_class'.format(bug_type)] = 'tab sel'
        if bug_type == 'target':
            template = self.template.get_template('target_index.html')
            template_info['targets'] = [
                (target, sorted(libbe.command.depend.get_blocked_by(
                            self.bugdirs, target)))
                for target in bugs]
        else:
            template = self.template.get_template('standard_index.html')           
        content = template.render(template_info)+'\n'
        return self.ok_response(
            environ, start_response, content, content_type='text/html')

    def bug(self, environ, start_response):
        try:
            bugdir_id,bug_id = environ['be-server.url_args']
        except:
            raise libbe.util.wsgi.HandlerError(404, 'Not Found')
        user_id = '{0}/{1}'.format(bugdir_id, bug_id)
        bugdir,bug,comment = (
            libbe.command.util.bugdir_bug_comment_from_user_id(
                self.bugdirs, user_id))
        if self.logger:
            self.logger.log(
                self.log_level, 'generate bug file for {0}/{1}'.format(
                    bugdir.uuid, bug.uuid))
        if bug.severity == 'target':
            index_type = 'target'
        elif bug.active:
            index_type = 'active'
        else:
            index_type = 'inactive'
        target = libbe.command.target.bug_target(self.bugdirs, bug)
        if target == bug:  # e.g. when bug.severity == 'target'
            target = None
        up_link = '../../{0}?type={1}'.format(self._index_file, index_type)
        bug.load_comments(load_full=True)
        bug.comment_root.sort(cmp=libbe.comment.cmp_time, reverse=True)
        template_info = {
            'title': self.title,
            'charset': 'UTF-8',
            'stylesheet': '../../style.css',
            'header': self.header,
            'backlinks': self.template.get_template('bug_backlinks.html'),
            'up_link': up_link,
            'index_type': index_type.capitalize(),
            'index_file': self._index_file,
            'bug': bug,
            'target': target,
            'comment_entry': self.template.get_template(
                'bug_comment_entry.html'),
            'comments': [(depth,comment) for depth,comment
                         in bug.comment_root.thread(flatten=False)],
            'bug_dir': self.bug_dir,
            'comment_dir': self._truncated_comment_id,
            'format_body': self._format_comment_body,
            'div_close': _DivCloser(),
            'strip_email': self._strip_email,
            'generation_time': self._generation_time(),
            }
        template = self.template.get_template('bug.html')
        content = template.render(template_info)
        return self.ok_response(
            environ, start_response, content, content_type='text/html')

    # helper functions
    def refresh(self):
        if time.time() > self._refresh:
            if self.logger:
                self.logger.log(self.log_level, 'refresh bugdirs')
            for bugdir in self.bugdirs.values():
                bugdir.load_all_bugs()
            self._refresh = time.time() + 60

    def _truncated_bugdir_id(self, bugdir):
        return libbe.util.id._truncate(
            bugdir.uuid, self.bugdirs.keys(),
            min_length=self.min_id_length)

    def _truncated_bug_id(self, bug):
        return libbe.util.id._truncate(
            bug.uuid, bug.sibling_uuids(),
            min_length=self.min_id_length)

    def _truncated_comment_id(self, comment):
        return libbe.util.id._truncate(
            comment.uuid, comment.sibling_uuids(),
            min_length=self.min_id_length)

    def bug_dir(self, bug):
        return '{0}/{1}'.format(
            self._truncated_bugdir_id(bug.bugdir),
            self._truncated_bug_id(bug))

    def _long_to_linked_user(self, text):
        """
        >>> import libbe.bugdir
        >>> bugdir = libbe.bugdir.SimpleBugDir(memory=False)
        >>> a = ServerApp(bugdirs={bugdir.uuid: bugdir})
        >>> a._long_to_linked_user('A link #abc123/a#, and a non-link #x#y#.')
        'A link <a href="./a/">abc/a</a>, and a non-link #x#y#.'
        >>> bugdir.cleanup()
        """
        replacer = libbe.util.id.IDreplacer(
            self.bugdirs, self._long_to_linked_user_replacer, wrap=False)
        return re.sub(
            libbe.util.id.REGEXP, replacer, text)

    def _long_to_linked_user_replacer(self, bugdirs, long_id):
        """
        >>> import libbe.bugdir
        >>> import libbe.util.id
        >>> bugdir = libbe.bugdir.SimpleBugDir(memory=False)
        >>> bugdirs = {bugdir.uuid: bugdir}
        >>> a = bugdir.bug_from_uuid('a')
        >>> uuid_gen = libbe.util.id.uuid_gen
        >>> libbe.util.id.uuid_gen = lambda : '0123'
        >>> c = a.new_comment('comment for link testing')
        >>> libbe.util.id.uuid_gen = uuid_gen
        >>> c.uuid
        '0123'
        >>> a = ServerApp(bugdirs=bugdirs)
        >>> a._long_to_linked_user_replacer(bugdirs, 'abc123')
        '#abc123#'
        >>> a._long_to_linked_user_replacer(bugdirs, 'abc123/a')
        '<a href="./a/">abc/a</a>'
        >>> a._long_to_linked_user_replacer(bugdirs, 'abc123/a/0123')
        '<a href="./a/#0123">abc/a/012</a>'
        >>> a._long_to_linked_user_replacer(bugdirs, 'x')
        '#x#'
        >>> a._long_to_linked_user_replacer(bugdirs, '')
        '##'
        >>> bugdir.cleanup()
        """
        try:
            p = libbe.util.id.parse_user(bugdirs, long_id)
        except (libbe.util.id.MultipleIDMatches,
                libbe.util.id.NoIDMatches,
                libbe.util.id.InvalidIDStructure), e:
            return '#%s#' % long_id # re-wrap failures
        if p['type'] == 'bugdir':
            return '#%s#' % long_id
        elif p['type'] == 'bug':
            bugdir,bug,comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, long_id))
            return '<a href="./%s/">%s</a>' \
                % (self._truncated_bug_id(bug), bug.id.user())
        elif p['type'] == 'comment':
            bugdir,bug,comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, long_id))
            return '<a href="./%s/#%s">%s</a>' \
                % (self._truncated_bug_id(bug),
                   self._truncated_comment_id(comment),
                   comment.id.user())
        raise Exception('Invalid id type %s for "%s"'
                        % (p['type'], long_id))

    def _format_comment_body(self, bug, comment):
        link_long_ids = False
        save_body = False
        value = comment.body
        if comment.content_type == 'text/html':
            link_long_ids = True
        elif comment.content_type.startswith('text/'):
            value = '<pre>\n'+self._escape(value)+'\n</pre>'
            link_long_ids = True
        elif comment.content_type.startswith('image/'):
            save_body = True
            value = '<img src="./%s/%s" />' % (
                self._truncated_bug_id(bug),
                self._truncated_comment_id(comment))
        else:
            save_body = True
            value = '<a href="./%s/%s">Link to %s file</a>.' % (
                self._truncated_bug_id(bug),
                self._truncated_comment_id(comment),
                comment.content_type)
        if link_long_ids == True:
            value = self._long_to_linked_user(value)
        if save_body == True:
            per_bug_dir = os.path.join(self.out_dir_bugs, bug.uuid)
            if not os.path.exists(per_bug_dir):
                os.mkdir(per_bug_dir)
            comment_path = os.path.join(per_bug_dir, comment.uuid)
            self._write_file(
                '<Files %s>\n  ForceType %s\n</Files>' \
                    % (comment.uuid, comment.content_type),
                [per_bug_dir, '.htaccess'], mode='a')
            self._write_file(comment.body,
                             [per_bug_dir, comment.uuid], mode='wb')
        return value

    def _generation_time(self):
        if self.generation_time:
            return self.generation_time
        return time.ctime()

    def _escape(self, string):
        if string == None:
            return ''
        return xml.sax.saxutils.escape(string)

    def _strip_email(self, string):
        if self.strip_email:
            name,address = email.utils.parseaddr(string)
            if name:
                return name
            return address
        return string

    def _load_templates(self, template_dir=None):
        if template_dir is not None:
            template_dir = os.path.abspath(os.path.expanduser(template_dir))

        self.template_dict = {
##
            'style.css':
"""body {
  font-family: "lucida grande", "sans serif";
  font-size: 14px;
  color: #333;
  width: auto;
  margin: auto;
}

div.main {
  padding: 20px;
  margin: auto;
  padding-top: 0;
  margin-top: 1em;
  background-color: #fcfcfc;
  border-radius: 10px;
  
}

div.footer {
  font-size: small;
  padding-left: 20px;
  padding-right: 20px;
  padding-top: 5px;
  padding-bottom: 5px;
  margin: auto;
  background: #305275;
  color: #fffee7;
  border-radius: 10px;
}

div.header {
    font-size: xx-large;
    padding-left: 20px;
    padding-right: 20px;
    padding-top: 10px;
    font-weight:bold;
    padding-bottom: 10px;
    background: #305275;
    color: #fffee7;
    border-radius: 10px;
}

th.target_name {
    text-align:left;
    border: 1px solid;
    border-color: #305275;
    background-color: #305275;
    color: #fff;
    width: auto;
    border-top-left-radius: 8px;
    border-top-right-radius: 8px;
    padding-left: 5px;
    padding-right: 5px;
}

table {
  border-style: solid;
  border: 1px #c3d9ff;
  border-spacing: 0px 0px;
  width: auto;
  padding: 0px;
  
  }

tb { border: 1px; }

tr {
  vertical-align: top;
  border: 1px #c3d9ff;
  border-style: dotted;
  width: auto;
  padding: 0px;
}

th {
    border-width: 1px;
    border-style: solid;
    border-color: #c3d9ff;
    border-collapse: collapse;
    padding-left: 5px;
    padding-right: 5px;
}


td {
    border-width: 1px;
    border-color: #c3d9ff;
    border-collapse: collapse;
    padding-left: 5px;
    padding-right: 5px;
    width: auto;
}

img { border-style: none; }

ul {
  list-style-type: none;
  padding: 0;
}

p { width: auto; }

p.backlink {
  width: auto;
  font-weight: bold;
}

a {
  background: inherit;
  text-decoration: none;
}

a { color: #553d41; }
a:hover { color: #003d41; }
a:visited { color: #305275; }
.footer a { color: #508d91; }

/* bug index pages */

td.tab {
  padding-right: 1em;
  padding-left: 1em;
}

td.sel.tab {
    background-color: #c3d9ff ;
    border: 1px solid #c3d9ff;
    font-weight:bold;    
    border-top-left-radius: 15px;
    border-top-right-radius: 15px;
}

td.nsel.tab { 
    border: 1px solid #c3d9ff;
    font-weight:bold;    
    border-top-left-radius: 5px;
    border-top-right-radius: 5px;
}

table.bug_list {
    border-width: 1px;
    border-style: solid;
    border-color: #c3d9ff;
    padding: 0px;
    width: 100%;            
    border: 1px solid #c3d9ff;
}

table.target_list {
    padding: 0px;
    width: 100%;
    margin-bottom: 10px;
}

table.target_list.td {
    border-width: 1px;
}

tr.wishlist { background-color: #DCFAFF;}
tr.wishlist:hover { background-color: #C2DCE1; }

tr.minor { background-color: #FFFFA6; }
tr.minor:hover { background-color: #E6E696; }

tr.serious { background-color: #FF9077;}
tr.serious:hover { background-color: #E6826B; }

tr.critical { background-color: #FF752A; }
tr.critical:hover { background-color: #D63905;}

tr.fatal { background-color: #FF3300;}
tr.fatal:hover { background-color: #D60000;}

td.uuid { width: 5%; border-style: dotted;}
td.status { width: 5%; border-style: dotted;}
td.severity { width: 5%; border-style: dotted;}
td.summary { border-style: dotted;}
td.date { width: 25%; border-style: dotted;}

/* bug detail pages */

td.bug_detail_label { text-align: right; border: none;}
td.bug_detail { border: none;}
td.bug_comment_label { text-align: right; vertical-align: top; }
td.bug_comment { }

div.comment {
  padding: 20px;
  padding-top: 20px;
  margin: auto;
  margin-top: 0;
}

div.root.comment {
  padding: 0px;
  /* padding-top: 0px; */
  padding-bottom: 20px;
}
""",
##
            'base.html':
"""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
  <head>
    <title>{{ title }}</title>
    <meta http-equiv="Content-Type" content="text/html; charset={{ charset }}" />
    <link rel="stylesheet" href="{{ stylesheet }}" type="text/css" />
  </head>
  <body>
    <div class="header">{{ header }}</div>
    <div class="main">
      {% block content %}{% endblock %}
    </div>
    <div class="footer">
      <p>Generated by <a href="http://www.bugseverywhere.org/">
      Bugs Everywhere</a> on {{ generation_time }}</p>
      <p>
        <a href="http://validator.w3.org/check?uri=referer">
          Validate XHTML</a>&nbsp;|&nbsp;
        <a href="http://jigsaw.w3.org/css-validator/check?uri=referer">
          Validate CSS</a>
      </p>
    </div>
  </body>
</html>
""",
##
            'bugdirs.html':
"""{% extends "base.html" %}

{% block content %}
{% if bugdirss %}
{% block bugdir_table %}{% endblock %}
{% else %}
<p>No bugdirs.</p>
{% endif %}
{% endblock %}
""",
##
            'index.html':
"""{% extends "base.html" %}

{% block content %}
<table>
  <tbody>
    <tr>
      <td class="{{ active_class }}"><a href="{% if index_file %}{{ index_file }}{% else %}.{% endif %}">Active Bugs</a></td>
      <td class="{{ inactive_class }}"><a href="{{ index_file }}?type=inactive">Inactive Bugs</a></td>
      <td class="{{ target_class }}"><a href="{{ index_file }}?type=target">Divided by target</a></td>
    </tr>
  </tbody>
</table>
{% if bugs %}
{% block bug_table %}{% endblock %}
{% else %}
<p>No bugs.</p>
{% endif %}
{% endblock %}
""",
##
            'standard_index.html':
"""{% extends "index.html" %}

{% block bug_table %}
<table class="bug_list">
  <thead>
    <tr>
      <th>UUID</th>
      <th>Status</th>
      <th>Severity</th>
      <th>Summary</th>
      <th>Date</th>
    </tr>
  </thead>
  <tbody>
    {% for bug in bugs %}
    {{ bug_entry.render({'bug':bug, 'dir':bug_dir(bug), 'index_file':index_file}) }}
    {% endfor %}
  </tbody>
</table>
{% endblock %}
""",
##
        'target_index.html':
"""{% extends "index.html" %}

{% block bug_table %}
{% for target,bugs in targets %}
<table class="target_list">
  <thead>
    <tr>
      <th class="target_name" colspan="5">
        Target: {{ target.summary|e }} ({{ target.status|e }})
      </th>
    </tr>
    <tr>
      <th>UUID</th>
      <th>Status</th>
      <th>Severity</th>
      <th>Summary</th>
      <th>Date</th>
    </tr>
  </thead>
  <tbody>
    {% for bug in bugs %}
    {{ bug_entry.render({'bug':bug, 'dir':bug_dir(bug), 'index_file':index_file}) }}
    {% endfor %}
  </tbody>
</table>
{% endfor %}
{% endblock %}
""",
##
            'index_bug_entry.html':
"""<tr class="{{ bug.severity }}">
  <td class="uuid"><a href="{{ dir }}/{{ index_file }}">{{ bug.id.user()|e }}</a></td>
  <td class="status"><a href="{{ dir }}/{{ index_file }}">{{ bug.status|e }}</a></td>
  <td class="severity"><a href="{{ dir }}/{{ index_file }}">{{ bug.severity|e }}</a></td>
  <td class="summary"><a href="{{ dir }}/{{ index_file }}">{{ bug.summary|e }}</a></td>
  <td class="date"><a href="{{ dir }}/{{ index_file }}">{{ (bug.time_string or '')|e }}</a></td>
</tr>
""",
##
        'bug.html':
"""{% extends "base.html" %}

{% block content %}
{{ backlinks.render({'up_link': up_link, 'index_type':index_type, 'index_file':index_file}) }}
<h1>Bug: {{ bug.id.user()|e }}</h1>

<table>
  <tbody>
    <tr><td class="bug_detail_label">ID :</td>
        <td class="bug_detail">{{ bug.uuid|e }}</td></tr>
    <tr><td class="bug_detail_label">Short name :</td>
        <td class="bug_detail">{{ bug.id.user()|e }}</td></tr>
    <tr><td class="bug_detail_label">Status :</td>
        <td class="bug_detail">{{ bug.status|e }}</td></tr>
    <tr><td class="bug_detail_label">Severity :</td>
        <td class="bug_detail">{{ bug.severity|e }}</td></tr>
    <tr><td class="bug_detail_label">Assigned :</td>
        <td class="bug_detail">{{ strip_email(bug.assigned or '')|e }}</td></tr>
    <tr><td class="bug_detail_label">Reporter :</td>
        <td class="bug_detail">{{ strip_email(bug.reporter or '')|e }}</td></tr>
    <tr><td class="bug_detail_label">Creator :</td>
        <td class="bug_detail">{{ strip_email(bug.creator or '')|e }}</td></tr>
    <tr><td class="bug_detail_label">Created :</td>
        <td class="bug_detail">{{ (bug.time_string or '')|e }}</td></tr>
{% if target %}
    <tr><td class="bug_detail_label">Target :</td>
        <td class="bug_detail"><a href="../../{{ bug_dir(target) }}/{{ index_file }}">{{ target.summary }}</a></td></tr>
{% endif %}
    <tr><td class="bug_detail_label">Summary :</td>
        <td class="bug_detail">{{ bug.summary|e }}</td></tr>
  </tbody>
</table>

<hr/>

{% if comments %}
{% for depth,comment in comments %}
{% if depth == 0 %}
<div class="comment root" id="C{{ comment_dir(comment) }}">
{% else %}
<div class="comment" id="C{{ comment_dir(comment) }}">
{% endif %}
{{ comment_entry.render({
       'depth':depth, 'bug': bug, 'comment':comment, 'comment_dir':comment_dir,
       'format_body': format_body, 'div_close': div_close,
       'strip_email': strip_email}) }}
{{ div_close(depth) }}
{% endfor %}
{% if comments[-1][0] > 0 %}
{{ div_close(0) }}
{% endif %}
{% else %}
<p>No comments.</p>
{% endif %}
{{ backlinks.render({'up_link': up_link, 'index_type': index_type}) }}
{% endblock %}
""",
##
            'bug_backlinks.html':
"""<p class="backlink"><a href="{{ up_link }}">Back to {{ index_type }} Index</a></p>
<p class="backlink"><a href="../../{{ index_file }}?type=target">Back to Target Index</a></p>
""",
##
            'bug_comment_entry.html':
"""<table>
  <tbody>
    <tr>
      <td class="bug_comment_label">Comment:</td>
      <td class="bug_comment">
        --------- Comment ---------<br/>
        ID: {{ comment.uuid }}<br/>
        Short name: {{ comment.id.user() }}<br/>
        From: {{ strip_email(comment.author or '')|e }}<br/>
        Date: {{ (comment.date or '')|e }}<br/>
        <br/>
        {{ format_body(bug, comment) }}
      </td>
    </tr>
  </tbody>
</table>
""",
            }

        loader = DictLoader(self.template_dict)

        if template_dir:
            file_system_loader = FileSystemLoader(template_dir)
            loader = ChoiceLoader([file_system_loader, loader])
        self.template = Environment(loader=loader)


class HTML (libbe.util.wsgi.ServerCommand):
    """Serve or dump browsable HTML for the current repository

    >>> import sys
    >>> import libbe.bugdir
    >>> bugdir = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bugdir.storage)
    >>> cmd = HTML(ui=ui)

    >>> export_path = os.path.join(bugdir.storage.repo, 'html_export')
    >>> ret = ui.run(cmd, {'output': export_path, 'export-html': True})
    >>> os.path.exists(export_path)
    True
    >>> os.path.exists(os.path.join(export_path, 'index.html'))
    True
    >>> os.path.exists(os.path.join(export_path, 'index_inactive.html'))
    True
    >>> os.path.exists(os.path.join(export_path, bugdir.uuid))
    True
    >>> for bug in sorted(bugdir):
    ...     if os.path.exists(os.path.join(
    ...             export_path, bugdir.uuid, bug.uuid, 'index.html')):
    ...         print('got {0}'.format(bug.uuid))
    ...     else:
    ...         print('missing {0}'.format(bug.uuid))
    got a
    got b

    >>> ui.cleanup()
    >>> bugdir.cleanup()
    """
    name = 'html'

    def __init__(self, *args, **kwargs):
        super(HTML, self).__init__(*args, **kwargs)
        # ServerApp cannot write, so drop some security options
        self.options = [
            option for option in self.options
            if option.name not in [
                'read-only',
                'notify',
                'auth',
                ]]

        self.options.extend([
                libbe.command.Option(name='template-dir', short_name='t',
                    help=('Use different templates.  Defaults to internal '
                          'templates'),
                    arg=libbe.command.Argument(
                        name='template-dir', metavar='DIR',
                        completion_callback=libbe.command.util.complete_path)),
                libbe.command.Option(name='title',
                    help='Set the bug repository title',
                    arg=libbe.command.Argument(
                        name='title', metavar='STRING',
                        default='Bugs Everywhere Issue Tracker')),
                libbe.command.Option(name='index-header',
                    help='Set the index page headers',
                    arg=libbe.command.Argument(
                        name='index-header', metavar='STRING',
                        default='Bugs Everywhere Bug List')),
                libbe.command.Option(name='min-id-length', short_name='l',
                    help=('Attempt to truncate bug and comment IDs to this '
                          'length.  Set to -1 for non-truncated IDs'),
                    arg=libbe.command.Argument(
                        name='min-id-length', metavar='INT',
                        default=-1, type='int')),
                libbe.command.Option(name='strip-email',
                    help='Strip email addresses from person fields.'),
                libbe.command.Option(name='export-html', short_name='e',
                    help='Export all HTML pages and exit.'),
                libbe.command.Option(name='output', short_name='o',
                    help='Set the output path for HTML export',
                    arg=libbe.command.Argument(
                        name='output', metavar='DIR', default='./html_export',
                        completion_callback=libbe.command.util.complete_path)),
                libbe.command.Option(name='export-template', short_name='E',
                    help='Export the default template and exit.'),
                libbe.command.Option(name='export-template-dir', short_name='d',
                    help='Set the directory for the template export',
                    arg=libbe.command.Argument(
                        name='export-template-dir', metavar='DIR',
                        default='./default-templates/',
                        completion_callback=libbe.command.util.complete_path)),
                ])

    def _run(self, **params):
        if True in [params['export-template'], params['export-html']]:
            app = self._get_app(
                logger=None, storage=None, index_file='index.html',
                generation_time=time.ctime(), **params)
            if params['export-template']:
                self._write_default_template(
                    template_dict=app.template_dict,
                    out_dir=params['export-template-dir'])
            elif params['export-html']:
                self._write_static_pages(app=app, out_dir=params['output'])
            return 0
        # provide defaults for the dropped options
        params['read-only'] = True
        params['notify'] = None
        params['auth'] = None
        return super(HTML, self)._run(**params)

    def _get_app(self, logger, storage, index_file='', generation_time=None,
                 **kwargs):
        return ServerApp(
            logger=logger, bugdirs=self._get_bugdirs(),
            template_dir=kwargs['template-dir'],
            title=kwargs['title'],
            header=kwargs['index-header'],
            index_file=index_file,
            min_id_length=kwargs['min-id-length'],
            strip_email=kwargs['strip-email'],
            generation_time=generation_time)

    def _long_help(self):
        return """
Example usage::

    $ be html

Then point your browser at ``http://localhost:8000/``.

If either ``--export-html`` or ``export-template`` is set, the command
will exit after the dump without serving anything over the wire.
"""

    def _write_default_template(self, template_dict, out_dir):
        out_dir = self._make_dir(out_dir)
        for filename,text in template_dict.iteritems():
            self._write_file(text, [out_dir, filename])

    def _write_static_pages(self, app, out_dir):
        url_mappings = [
            ('index.html?type=active', 'index.html'),
            ('index.html?type=inactive', 'index_inactive.html'),
            ('index.html?type=target', 'index_by_target.html'),
            ]
        out_dir = self._make_dir(out_dir)
        caller = libbe.util.wsgi.WSGICaller()
        self._write_file(
            content=self._get_content(caller, app, 'style.css'),
            path_array=[out_dir, 'style.css'])
        for url,data_dict,path in [
            ('index.html', {'type': 'active'}, 'index.html'),
            ('index.html', {'type': 'inactive'}, 'index_inactive.html'),
            ('index.html', {'type': 'target'}, 'index_by_target.html'),
            ]:
            content = self._get_content(caller, app, url, data_dict)
            for url_,path_ in url_mappings:
                content = content.replace(url_, path_)
            self._write_file(content=content, path_array=[out_dir, path])
        for bugdir in app.bugdirs.values():
            for bug in bugdir:
                bug_dir_url = app.bug_dir(bug=bug)
                segments = bug_dir_url.split('/')
                path_array = [out_dir]
                path_array.extend(segments)
                bug_dir_path = os.path.join(*path_array)
                path_array.append(app._index_file)
                url = '{0}/{1}'.format(bug_dir_url, app._index_file)
                content = self._get_content(caller, app, url)
                for url_,path_ in url_mappings:
                    content = content.replace(url_, path_)
                if not os.path.isdir(bug_dir_path):
                    self._make_dir(bug_dir_path)                    
                self._write_file(content=content, path_array=path_array)

    def _get_content(self, caller, app, path, data_dict=None):
        try:
            return caller.getURL(app=app, path=path, data_dict=data_dict)
        except libbe.util.wsgi.HandlerError:
            self.stdout.write(
                'error retrieving {0} with {1}\n'.format(path, data_dict))
            raise

    def _make_dir(self, dir_path):
        dir_path = os.path.abspath(os.path.expanduser(dir_path))
        if not os.path.exists(dir_path):
            try:
                os.makedirs(dir_path)
            except:
                raise libbe.command.UserError(
                    'Cannot create output directory "{0}".'.format(dir_path))
        return dir_path

    def _write_file(self, content, path_array, mode='w'):
        if not hasattr(self, 'encoding'):
            self.encoding = libbe.util.encoding.get_text_file_encoding()
        return libbe.util.encoding.set_file_contents(
            os.path.join(*path_array), content, mode, self.encoding)


Html = HTML # alias for libbe.command.base.get_command_class()


class _DivCloser (object):
    def __init__(self, depth=0):
        self.depth = depth

    def __call__(self, depth):
        ret = []
        while self.depth >= depth:
            self.depth -= 1
            ret.append('</div>')
        self.depth = depth
        return '\n'.join(ret)



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Serializing and deserializing dictionaries of parameters.

The serialized "mapfiles" should be clear, flat-text strings, and allow
easy merging of independent/conflicting changes.
"""

import errno
import json
import os.path

import libbe
if libbe.TESTING == True:
    import doctest


class InvalidMapfileContents (Exception):
    def __init__(self, contents):
        super(InvalidMapfileContents, self).__init__('Invalid JSON contents')
        self.contents = contents


def generate(map, context=6):
    """Generate a JSON mapfile content string.

    Examples
    --------

    >>> import sys
    >>> sys.stdout.write(generate({}))
    {}
    >>> sys.stdout.write(generate({'q':'p'}))
    {
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
        "q": "p"
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
    <BLANKLINE>
    }

    The blank lines ensure that merging occurs independent of
    surrounding content.  Because the mapfile format is also used by
    the :py:mod:`~libbe.command.serve_commands` where merging is not
    important, the amount of context is controllable.

    >>> sys.stdout.write(generate({'q':u'Fran\u00e7ais'}, context=0))
    {
        "q": "Fran\\u00e7ais"
    }
    >>> sys.stdout.write(generate({'q':u'hello'}, context=0))
    {
        "q": "hello"
    }
    >>> sys.stdout.write(generate(
    ...         {'p':'really long line\\n'*10, 'q': 'the next entry'},
    ...         context=1))
    {
    <BLANKLINE>
        "p": "really long line\\nreally long line\\nreally long line\\nreally long line\\nreally long line\\nreally long line\\nreally long line\\nreally long line\\nreally long line\\nreally long line\\n", 
    <BLANKLINE>
        "q": "the next entry"
    <BLANKLINE>
    }

    See Also
    --------
    parse : inverse
    """
    lines = json.dumps(map, sort_keys=True, indent=4).splitlines()
    sep = '\n' * (1 + context)
    return sep.join(lines) + '\n'

def parse(contents):
    """Parse a JSON mapfile string.

    Examples
    --------

    >>> parse('{"q": "p"}')['q']
    u'p'
    >>> contents = generate({'a':'b', 'c':'d', 'e':'f'})
    >>> dict = parse(contents)
    >>> dict['a']
    u'b'
    >>> dict['c']
    u'd'
    >>> dict['e']
    u'f'
    >>> contents = generate({'q':u'Fran\u00e7ais'})
    >>> dict = parse(contents)
    >>> dict['q']
    u'Fran\\xe7ais'
    >>> dict = parse('a!')
    Traceback (most recent call last):
      ...
    InvalidMapfileContents: Invalid JSON contents

    See Also
    --------
    generate : inverse

    """
    try:
        return json.loads(contents)
    except ValueError:
        raise InvalidMapfileContents(contents)

if libbe.TESTING == True:
    suite = doctest.DocTestSuite()



# Bugs Everywhere, a distributed bugtracker
# Copyright (C) 2008-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define :py:class:`Tree`, a traversable tree structure.
"""

import libbe
if libbe.TESTING == True:
    import doctest

class Tree(list):
    """A traversable tree structure.

    Examples
    --------

    Construct::

               +-b---d-g
             a-+   +-e
               +-c-+-f-h-i

    with

    >>> i = Tree();       i.n = "i"
    >>> h = Tree([i]);    h.n = "h"
    >>> f = Tree([h]);    f.n = "f"
    >>> e = Tree();       e.n = "e"
    >>> c = Tree([f,e]);  c.n = "c"
    >>> g = Tree();       g.n = "g"
    >>> d = Tree([g]);    d.n = "d"
    >>> b = Tree([d]);    b.n = "b"
    >>> a = Tree();       a.n = "a"
    >>> a.append(c)
    >>> a.append(b)

    Get the longest branch length with

    >>> a.branch_len()
    5

    Sort the tree recursively.  Here we sort longest branch length
    first.

    >>> a.sort(key=lambda node : -node.branch_len())
    >>> "".join([node.n for node in a.traverse()])
    'acfhiebdg'

    And here we sort shortest branch length first.

    >>> a.sort(key=lambda node : node.branch_len())
    >>> "".join([node.n for node in a.traverse()])
    'abdgcefhi'

    We can also do breadth-first traverses.

    >>> "".join([node.n for node in a.traverse(depth_first=False)])
    'abcdefghi'

    Serialize the tree with depth marking branches.

    >>> for depth,node in a.thread():
    ...     print "%*s" % (2*depth+1, node.n)
    a
      b
        d
          g
      c
        e
        f
          h
            i

    Flattening the thread disables depth increases except at
    branch splits.

    >>> for depth,node in a.thread(flatten=True):
    ...     print "%*s" % (2*depth+1, node.n)
    a
      b
      d
      g
    c
      e
    f
    h
    i

    We can also check if a node is contained in a tree.

    >>> a.has_descendant(g)
    True
    >>> c.has_descendant(g)
    False
    >>> a.has_descendant(a)
    False
    >>> a.has_descendant(a, match_self=True)
    True
    """
    def __cmp__(self, other):
        return cmp(id(self), id(other))

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def branch_len(self):
        """Return the largest number of nodes from root to leaf (inclusive).

        For the tree::

               +-b---d-g
             a-+   +-e
               +-c-+-f-h-i

        this method returns 5.

        Notes
        -----
        Exhaustive search every time == *slow*.

        Use only on small trees, or reimplement by overriding
        child-addition methods to allow accurate caching.
        """
        if len(self) == 0:
            return 1
        else:
            return 1 + max([child.branch_len() for child in self])

    def sort(self, *args, **kwargs):
        """Sort the tree recursively.

        This method extends :py:meth:`list.sort` to Trees.

        Notes
        -----
        This method can be slow, e.g. on a :py:meth:`branch_len` sort,
        since a node at depth `N` from the root has it's
        :py:meth:`branch_len` method called `N` times.
        """
        list.sort(self, *args, **kwargs)
        for child in self:
            child.sort(*args, **kwargs)

    def traverse(self, depth_first=True):
        """Generate all the nodes in a tree, starting with the root node.

        Parameters
        ----------
        depth_first : bool
          Depth first by default, but you can set `depth_first` to
          `False` for breadth first ordering.  Siblings are returned
          in the order they are stored, so you might want to
          :py:meth:`sort` your tree first.
        """
        if depth_first == True:
            yield self
            for child in self:
                for descendant in child.traverse():
                    yield descendant
        else: # breadth first, Wikipedia algorithm
            # http://en.wikipedia.org/wiki/Breadth-first_search
            queue = [self]
            while len(queue) > 0:
                node = queue.pop(0)
                yield node
                queue.extend(node)

    def thread(self, flatten=False):
        """Generate a (depth, node) tuple for every node in the tree.

        When `flatten` is `False`, the depth of any node is one
        greater than the depth of its parent.  That way the
        inheritance is explicit, but you can end up with highly
        indented threads.

        When `flatten` is `True`, the depth of any node is only
        greater than the depth of its parent when there is a branch,
        and the node is not the last child.  This can lead to ancestry
        ambiguity, but keeps the total indentation down.  For example::

                      +-b                  +-b-c
                    a-+-c        and     a-+
                      +-d-e-f              +-d-e-f

        would both produce (after sorting by :py:meth:`branch_len`)::

            (0, a)
            (1, b)
            (1, c)
            (0, d)
            (0, e)
            (0, f)

        """
        stack = [] # ancestry of the current node
        if flatten == True:
            depthDict = {}

        for node in self.traverse(depth_first=True):
            while len(stack) > 0 \
                    and id(node) not in [id(c) for c in stack[-1]]:
                stack.pop(-1)
            if flatten == False:
                depth = len(stack)
            else:
                if len(stack) == 0:
                    depth = 0
                else:
                    parent = stack[-1]
                    depth = depthDict[id(parent)]
                    if len(parent) > 1 and node != parent[-1]:
                        depth += 1
                depthDict[id(node)] = depth
            yield (depth,node)
            stack.append(node)

    def has_descendant(self, descendant, depth_first=True, match_self=False):
        """Check if a node is contained in a tree.

        Parameters
        ----------
        descendant : Tree
          The potential descendant.
        depth_first : bool
          The search order.  Set this if you feel depth/breadth would
          be a faster search.
        match_self : bool
          Set to `True` for::

              x.has_descendant(x, match_self=True) -> True
        """
        if descendant == self:
            return match_self
        for d in self.traverse(depth_first):
            if descendant == d:
                return True
        return False

if libbe.TESTING == True:
    suite = doctest.DocTestSuite()



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Niall Douglas (s_sourceforge@nedprod.com) <spam@spamtrap.com>
#                         Robert Lehmann <mail@robertlehmann.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

import libbe
import libbe.command
import libbe.command.util
import libbe.comment
import libbe.ui.util.editor
import libbe.util.id


class Comment (libbe.command.Command):
    """Add a comment to a bug

    >>> import time
    >>> import libbe.bugdir
    >>> import libbe.util.id
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Comment(ui=ui)

    >>> uuid_gen = libbe.util.id.uuid_gen
    >>> libbe.util.id.uuid_gen = lambda: 'X'
    >>> ui._user_id = u'Fran\\xe7ois'
    >>> ret = ui.run(cmd, args=['/a', 'This is a comment about a'])
    Created comment with ID abc/a/X
    >>> libbe.util.id.uuid_gen = uuid_gen
    >>> bd.flush_reload()
    >>> bug = bd.bug_from_uuid('a')
    >>> bug.load_comments(load_full=False)
    >>> comment = bug.comment_root[0]
    >>> comment.id.storage() == comment.uuid
    True
    >>> print comment.body
    This is a comment about a
    <BLANKLINE>
    >>> comment.author
    u'Fran\\xe7ois'
    >>> comment.time <= int(time.time())
    True
    >>> comment.in_reply_to is None
    True

    >>> if 'EDITOR' in os.environ:
    ...     del os.environ['EDITOR']
    >>> if 'VISUAL' in os.environ:
    ...     del os.environ['VISUAL']
    >>> ui._user_id = u'Frank'
    >>> ret = ui.run(cmd, args=['/b'])
    Traceback (most recent call last):
    UserError: No comment supplied, and EDITOR not specified.

    >>> os.environ['EDITOR'] = "echo 'I like cheese' > "
    >>> libbe.util.id.uuid_gen = lambda: 'Y'
    >>> ret = ui.run(cmd, args=['/b'])
    Created comment with ID abc/b/Y
    >>> libbe.util.id.uuid_gen = uuid_gen
    >>> bd.flush_reload()
    >>> bug = bd.bug_from_uuid('b')
    >>> bug.load_comments(load_full=False)
    >>> comment = bug.comment_root[0]
    >>> print comment.body
    I like cheese
    <BLANKLINE>
    >>> ui.cleanup()
    >>> bd.cleanup()
    >>> del os.environ["EDITOR"]
    """
    name = 'comment'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='author', short_name='a',
                    help='Set the comment author',
                    arg=libbe.command.Argument(
                        name='author', metavar='AUTHOR')),
                libbe.command.Option(name='alt-id',
                    help='Set an alternate comment ID',
                    arg=libbe.command.Argument(
                        name='alt-id', metavar='ID')),
                libbe.command.Option(name='content-type', short_name='c',
                    help='Set comment content-type (e.g. text/plain)',
                    arg=libbe.command.Argument(name='content-type',
                        metavar='MIME')),
                libbe.command.Option(name='full-uuid', short_name='f',
                    help='Print the full UUID for the new bug')
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='id', metavar='ID', default=None,
                    completion_callback=libbe.command.util.complete_bug_comment_id),
                libbe.command.Argument(
                    name='comment', metavar='COMMENT', default=None,
                    optional=True,
                    completion_callback=libbe.command.util.complete_assigned),
                ])

    def _run(self, **params):
        bugdirs = self._get_bugdirs()
        bugdir,bug,parent = (
            libbe.command.util.bugdir_bug_comment_from_user_id(
                bugdirs, params['id']))
        if params['comment'] == None:
            # try to launch an editor for comment-body entry
            try:
                if parent == bug.comment_root:
                    header = "Subject: %s" % bug.summary
                    parent_body = parent.string_thread() or "No comments"
                else:
                    header = "From: %s\nTo: %s" % (parent.author, bug)
                    parent_body = parent.body
                estr = 'Please enter your comment above\n\n%s\n\n> %s\n' \
                    % (header, '\n> '.join(parent_body.splitlines()))
                body = libbe.ui.util.editor.editor_string(estr)
            except libbe.ui.util.editor.CantFindEditor, e:
                raise libbe.command.UserError(
                    'No comment supplied, and EDITOR not specified.')
            if body is None:
                raise libbe.command.UserError('No comment entered.')
        elif params['comment'] == '-': # read body from stdin
            binary = not (params['content-type'] == None
                          or params['content-type'].startswith("text/"))
            if not binary:
                body = self.stdin.read()
                if not body.endswith('\n'):
                    body += '\n'
            else: # read-in without decoding
                body = sys.stdin.read()
        else: # body given on command line
            body = params['comment']
            if not body.endswith('\n'):
                body+='\n'
        if params['author'] == None:
            params['author'] = self._get_user_id()

        new = parent.new_reply(body=body, content_type=params['content-type'])
        for key in ['alt-id', 'author']:
            if params[key] != None:
                setattr(new, new._setting_name_to_attr_name(key), params[key])
        if params['full-uuid']:
            comment_id = new.id.long_user()
        else:
            comment_id = new.id.user()
        self.stdout.write('Created comment with ID %s\n' % (comment_id))
        return 0

    def _long_help(self):
        return """
To add a comment to a bug, use the bug ID as the argument.  To reply
to another comment, specify the comment name (as shown in "be show"
output).  COMMENT, if specified, should be either the text of your
comment or "-", in which case the text will be read from stdin.  If
you do not specify a COMMENT, $EDITOR is used to launch an editor.  If
COMMENT is unspecified and EDITOR is not set, no comment will be
created.
"""



#!/usr/bin/env python
# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Store version info for this BE installation.

By default, use the Git-generated information in
:py:mod:`~libbe._version`, but allow manual overriding by setting
:py:data:`libbe.version._VERSION`.  This allows support of both the "I
don't want to be bothered setting version strings" and the "I want
complete control over the version strings" workflows.
"""

import copy

import libbe.storage
try:
    from ._version import version_info    
except ImportError as e:
    import logging
    logging.warn('unable to import libbe._version: {0}'.format(e))
    version_info = {
        'revision': 'unknown',
        'date': 'unknown',
        'committer': 'unknown',
        }

# Manually set a version string (optional, defaults to bzr revision id)
_VERSION = '1.1.1'

def version(verbose=False):
    """
    Returns the version string for this BE installation.  If
    verbose==True, the string will include extra lines with more
    detail (e.g. last committer's name, etc.).
    """
    if "_VERSION" in globals():
        string = _VERSION
    else:
        string = version_info['revision'][:8]
    if verbose == True:
        info = copy.copy(version_info)
        info['storage'] = libbe.storage.STORAGE_VERSION
        string += (
            '\n'
            'revision: {revision}\n'
            'date: {date}\n'
            'committer: {committer}\n'
            'storage version: {storage}'
            ).format(**info)
    return string


if __name__ == '__main__':
    print(version(verbose=True))



# Bugs Everywhere - a distributed bugtracker
# Copyright (C) 2008-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Provides a series of useful decorators for defining various types
of properties.

For example usage, consider the unittests at the end of the module.

Notes
-----

See `PEP 318` and Michele Simionato's `decorator documentation` for
more information on decorators.

.. _PEP 318: http://www.python.org/dev/peps/pep-0318/
.. _decorator documentation: http://www.phyast.pitt.edu/~micheles/python/documentation.html

See Also
--------
:py:mod:`libbe.storage.util.settings_object` : bundle properties into a convenient package

"""

import copy
import types

import libbe
if libbe.TESTING == True:
    import unittest


class ValueCheckError (ValueError):
    def __init__(self, name, value, allowed):
        action = "in" # some list of allowed values
        if type(allowed) == types.FunctionType:
            action = "allowed by" # some allowed-value check function
        msg = "%s not %s %s for %s" % (value, action, allowed, name)
        ValueError.__init__(self, msg)
        self.name = name
        self.value = value
        self.allowed = allowed

def Property(funcs):
    """
    End a chain of property decorators, returning a property.
    """
    args = {}
    args["fget"] = funcs.get("fget", None)
    args["fset"] = funcs.get("fset", None)
    args["fdel"] = funcs.get("fdel", None)
    args["doc"] = funcs.get("doc", None)

    #print "Creating a property with"
    #for key, val in args.items(): print key, value
    return property(**args)

def doc_property(doc=None):
    """
    Add a docstring to a chain of property decorators.
    """
    def decorator(funcs=None):
        """
        Takes either a dict of funcs {"fget":fnX, "fset":fnY, ...}
        or a function fn() returning such a dict.
        """
        if hasattr(funcs, "__call__"):
            funcs = funcs() # convert from function-arg to dict
        funcs["doc"] = doc
        return funcs
    return decorator

def local_property(name, null=None, mutable_null=False):
    """
    Define get/set access to per-parent-instance local storage.  Uses
    ._<name>_value to store the value for a particular owner instance.
    If the ._<name>_value attribute does not exist, returns null.

    If mutable_null == True, we only release deepcopies of the null to
    the outside world.
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget", None)
        fset = funcs.get("fset", None)
        def _fget(self):
            if fget is not None:
                fget(self)
            if mutable_null == True:
                ret_null = copy.deepcopy(null)
            else:
                ret_null = null
            value = getattr(self, "_%s_value" % name, ret_null)
            return value
        def _fset(self, value):
            setattr(self, "_%s_value" % name, value)
            if fset is not None:
                fset(self, value)
        funcs["fget"] = _fget
        funcs["fset"] = _fset
        funcs["name"] = name
        return funcs
    return decorator

def settings_property(name, null=None):
    """
    Similar to local_property, except where local_property stores the
    value in instance._<name>_value, settings_property stores the
    value in instance.settings[name].
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget", None)
        fset = funcs.get("fset", None)
        def _fget(self):
            if fget is not None:
                fget(self)
            value = self.settings.get(name, null)
            return value
        def _fset(self, value):
            self.settings[name] = value
            if fset is not None:
                fset(self, value)
        funcs["fget"] = _fget
        funcs["fset"] = _fset
        funcs["name"] = name
        return funcs
    return decorator


# Allow comparison and caching with _original_ values for mutables,
# since
#
# >>> a = []
# >>> b = a
# >>> b.append(1)
# >>> a
# [1]
# >>> a==b
# True
def _hash_mutable_value(value):
    return repr(value)
def _init_mutable_property_cache(self):
    if not hasattr(self, "_mutable_property_cache_hash"):
        # first call to _fget for any mutable property
        self._mutable_property_cache_hash = {}
        self._mutable_property_cache_copy = {}
def _set_cached_mutable_property(self, cacher_name, property_name, value):
    _init_mutable_property_cache(self)
    self._mutable_property_cache_hash[(cacher_name, property_name)] = \
        _hash_mutable_value(value)
    self._mutable_property_cache_copy[(cacher_name, property_name)] = \
        copy.deepcopy(value)
def _get_cached_mutable_property(self, cacher_name, property_name, default=None):
    _init_mutable_property_cache(self)
    if (cacher_name, property_name) not in self._mutable_property_cache_copy:
        return default
    return self._mutable_property_cache_copy[(cacher_name, property_name)]
def _cmp_cached_mutable_property(self, cacher_name, property_name, value, default=None):
    _init_mutable_property_cache(self)
    if (cacher_name, property_name) not in self._mutable_property_cache_hash:
        _set_cached_mutable_property(self, cacher_name, property_name, default)
    old_hash = self._mutable_property_cache_hash[(cacher_name, property_name)]
    return cmp(_hash_mutable_value(value), old_hash)


def defaulting_property(default=None, null=None,
                        mutable_default=False):
    """
    Define a default value for get access to a property.
    If the stored value is null, then default is returned.

    If mutable_default == True, we only release deepcopies of the
    default to the outside world.

    null should never escape to the outside world, so don't worry
    about it being a mutable.
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget")
        fset = funcs.get("fset")
        name = funcs.get("name", "<unknown>")
        def _fget(self):
            value = fget(self)
            if value == null:
                if mutable_default == True:
                    return copy.deepcopy(default)
                else:
                    return default
            return value
        def _fset(self, value):
            if value == default:
                value = null
            fset(self, value)
        funcs["fget"] = _fget
        funcs["fset"] = _fset
        return funcs
    return decorator

def fn_checked_property(value_allowed_fn):
    """
    Define allowed values for get/set access to a property.
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget")
        fset = funcs.get("fset")
        name = funcs.get("name", "<unknown>")
        def _fget(self):
            value = fget(self)
            if value_allowed_fn(value) != True:
                raise ValueCheckError(name, value, value_allowed_fn)
            return value
        def _fset(self, value):
            if value_allowed_fn(value) != True:
                raise ValueCheckError(name, value, value_allowed_fn)
            fset(self, value)
        funcs["fget"] = _fget
        funcs["fset"] = _fset
        return funcs
    return decorator

def checked_property(allowed=[]):
    """
    Define allowed values for get/set access to a property.
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget")
        fset = funcs.get("fset")
        name = funcs.get("name", "<unknown>")
        def _fget(self):
            value = fget(self)
            if value not in allowed:
                raise ValueCheckError(name, value, allowed)
            return value
        def _fset(self, value):
            if value not in allowed:
                raise ValueCheckError(name, value, allowed)
            fset(self, value)
        funcs["fget"] = _fget
        funcs["fset"] = _fset
        return funcs
    return decorator

def cached_property(generator, initVal=None, mutable=False):
    """
    Allow caching of values generated by generator(instance), where
    instance is the instance to which this property belongs.  Uses
    ._<name>_cache to store a cache flag for a particular owner
    instance.

    When the cache flag is True or missing and the stored value is
    initVal, the first fget call triggers the generator function,
    whose output is stored in _<name>_cached_value.  That and
    subsequent calls to fget will return this cached value.

    If the input value is no longer initVal (e.g. a value has been
    loaded from disk or set with fset), that value overrides any
    cached value, and this property has no effect.

    When the cache flag is False and the stored value is initVal, the
    generator is not cached, but is called on every fget.

    The cache flag is missing on initialization.  Particular instances
    may override by setting their own flag.

    In the case that mutable == True, all caching is disabled and the
    generator is called whenever the cached value would otherwise be
    used.
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget")
        name = funcs.get("name", "<unknown>")
        def _fget(self):
            cache = getattr(self, "_%s_cache" % name, True)
            value = fget(self)
            if value == initVal:
                if cache == True and mutable == False:
                    if hasattr(self, "_%s_cached_value" % name):
                        value = getattr(self, "_%s_cached_value" % name)
                    else:
                        value = generator(self)
                        setattr(self, "_%s_cached_value" % name, value)
                else:
                    value = generator(self)
            return value
        funcs["fget"] = _fget
        return funcs
    return decorator

def primed_property(primer, initVal=None, unprimeableVal=None):
    """
    Just like a cached_property, except that instead of returning a
    new value and running fset to cache it, the primer attempts some
    background manipulation (e.g. loads data into instance.settings)
    such that a _second_ pass through fget succeeds.  If the second
    pass doesn't succeed (e.g. no readable storage), we give up and
    return unprimeableVal.

    The 'cache' flag becomes a 'prime' flag, with priming taking place
    whenever ._<name>_prime is True, or is False or missing and
    value == initVal.
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget")
        name = funcs.get("name", "<unknown>")
        def _fget(self):
            prime = getattr(self, "_%s_prime" % name, False)
            if prime == False:
                value = fget(self)
            if prime == True or (prime == False and value == initVal):
                primer(self)
                value = fget(self)
                if prime == False and value == initVal:
                    return unprimeableVal
            return value
        funcs["fget"] = _fget
        return funcs
    return decorator

def change_hook_property(hook, mutable=False, default=None):
    """Call the function `hook` whenever a value different from the
    current value is set.

    This is useful for saving changes to disk, etc.  This function is
    called *after* the new value has been stored, allowing you to
    change the stored value if you want.

    In the case of mutables, things are slightly trickier.  Because
    the property-owning class has no way of knowing when the value
    changes.  We work around this by caching a private deepcopy of the
    mutable value, and checking for changes whenever the property is
    set (obviously) or retrieved (to check for external changes).  So
    long as you're conscientious about accessing the property after
    making external modifications, mutability won't be a problem::

      t.x.append(5) # external modification
      t.x           # dummy access notices change and triggers hook

    See :py:class:`testChangeHookMutableProperty` for an example of the
    expected behavior.

    Parameters
    ----------
    hook : fn
      `hook(instance, old_value, new_value)`, where `instance` is a
      reference to the class instance to which this property belongs.
    """
    def decorator(funcs):
        if hasattr(funcs, "__call__"):
            funcs = funcs()
        fget = funcs.get("fget")
        fset = funcs.get("fset")
        name = funcs.get("name", "<unknown>")
        def _fget(self, new_value=None, from_fset=False): # only used if mutable == True
            if from_fset == True:
                value = new_value # compare new value with cached
            else:
                value = fget(self) # compare current value with cached
            if _cmp_cached_mutable_property(self, "change hook property", name, value, default) != 0:
                # there has been a change, cache new value
                old_value = _get_cached_mutable_property(self, "change hook property", name, default)
                _set_cached_mutable_property(self, "change hook property", name, value)
                if from_fset == True: # return previously cached value
                    value = old_value
                else: # the value changed while we weren't looking
                    hook(self, old_value, value)
            return value
        def _fset(self, value):
            if mutable == True: # get cached previous value
                old_value = _fget(self, new_value=value, from_fset=True)
            else:
                old_value = fget(self)
            fset(self, value)
            if value != old_value:
                hook(self, old_value, value)
        if mutable == True:
            funcs["fget"] = _fget
        funcs["fset"] = _fset
        return funcs
    return decorator

if libbe.TESTING == True:
    class DecoratorTests(unittest.TestCase):
        def testLocalDoc(self):
            class Test(object):
                @Property
                @doc_property("A fancy property")
                def x():
                    return {}
            self.failUnless(Test.x.__doc__ == "A fancy property",
                            Test.x.__doc__)
        def testLocalProperty(self):
            class Test(object):
                @Property
                @local_property(name="LOCAL")
                def x():
                    return {}
            t = Test()
            self.failUnless(t.x == None, str(t.x))
            t.x = 'z' # the first set initializes ._LOCAL_value
            self.failUnless(t.x == 'z', str(t.x))
            self.failUnless("_LOCAL_value" in dir(t), dir(t))
            self.failUnless(t._LOCAL_value == 'z', t._LOCAL_value)
        def testSettingsProperty(self):
            class Test(object):
                @Property
                @settings_property(name="attr")
                def x():
                    return {}
                def __init__(self):
                    self.settings = {}
            t = Test()
            self.failUnless(t.x == None, str(t.x))
            t.x = 'z' # the first set initializes ._LOCAL_value
            self.failUnless(t.x == 'z', str(t.x))
            self.failUnless("attr" in t.settings, t.settings)
            self.failUnless(t.settings["attr"] == 'z', t.settings["attr"])
        def testDefaultingLocalProperty(self):
            class Test(object):
                @Property
                @defaulting_property(default='y', null='x')
                @local_property(name="DEFAULT", null=5)
                def x(): return {}
            t = Test()
            self.failUnless(t.x == 5, str(t.x))
            t.x = 'x'
            self.failUnless(t.x == 'y', str(t.x))
            t.x = 'y'
            self.failUnless(t.x == 'y', str(t.x))
            t.x = 'z'
            self.failUnless(t.x == 'z', str(t.x))
            t.x = 5
            self.failUnless(t.x == 5, str(t.x))
        def testCheckedLocalProperty(self):
            class Test(object):
                @Property
                @checked_property(allowed=['x', 'y', 'z'])
                @local_property(name="CHECKED")
                def x(): return {}
                def __init__(self):
                    self._CHECKED_value = 'x'
            t = Test()
            self.failUnless(t.x == 'x', str(t.x))
            try:
                t.x = None
                e = None
            except ValueCheckError, e:
                pass
            self.failUnless(type(e) == ValueCheckError, type(e))
        def testTwoCheckedLocalProperties(self):
            class Test(object):
                @Property
                @checked_property(allowed=['x', 'y', 'z'])
                @local_property(name="X")
                def x(): return {}

                @Property
                @checked_property(allowed=['a', 'b', 'c'])
                @local_property(name="A")
                def a(): return {}
                def __init__(self):
                    self._A_value = 'a'
                    self._X_value = 'x'
            t = Test()
            try:
                t.x = 'a'
                e = None
            except ValueCheckError, e:
                pass
            self.failUnless(type(e) == ValueCheckError, type(e))
            t.x = 'x'
            t.x = 'y'
            t.x = 'z'
            try:
                t.a = 'x'
                e = None
            except ValueCheckError, e:
                pass
            self.failUnless(type(e) == ValueCheckError, type(e))
            t.a = 'a'
            t.a = 'b'
            t.a = 'c'
        def testFnCheckedLocalProperty(self):
            class Test(object):
                @Property
                @fn_checked_property(lambda v : v in ['x', 'y', 'z'])
                @local_property(name="CHECKED")
                def x(): return {}
                def __init__(self):
                    self._CHECKED_value = 'x'
            t = Test()
            self.failUnless(t.x == 'x', str(t.x))
            try:
                t.x = None
                e = None
            except ValueCheckError, e:
                pass
            self.failUnless(type(e) == ValueCheckError, type(e))
        def testCachedLocalProperty(self):
            class Gen(object):
                def __init__(self):
                    self.i = 0
                def __call__(self, owner):
                    self.i += 1
                    return self.i
            class Test(object):
                @Property
                @cached_property(generator=Gen(), initVal=None)
                @local_property(name="CACHED")
                def x(): return {}
            t = Test()
            self.failIf("_CACHED_cache" in dir(t),
                        getattr(t, "_CACHED_cache", None))
            self.failUnless(t.x == 1, t.x)
            self.failUnless(t.x == 1, t.x)
            self.failUnless(t.x == 1, t.x)
            t.x = 8
            self.failUnless(t.x == 8, t.x)
            self.failUnless(t.x == 8, t.x)
            t._CACHED_cache = False        # Caching is off, but the stored value
            val = t.x                      # is 8, not the initVal (None), so we
            self.failUnless(val == 8, val) # get 8.
            t._CACHED_value = None         # Now we've set the stored value to None
            val = t.x                      # so future calls to fget (like this)
            self.failUnless(val == 2, val) # will call the generator every time...
            val = t.x
            self.failUnless(val == 3, val)
            val = t.x
            self.failUnless(val == 4, val)
            t._CACHED_cache = True              # We turn caching back on, and get
            self.failUnless(t.x == 1, str(t.x)) # the original cached value.
            del t._CACHED_cached_value          # Removing that value forces a
            self.failUnless(t.x == 5, str(t.x)) # single cache-regenerating call
            self.failUnless(t.x == 5, str(t.x)) # to the genenerator, after which
            self.failUnless(t.x == 5, str(t.x)) # we get the new cached value.
        def testPrimedLocalProperty(self):
            class Test(object):
                def prime(self):
                    self.settings["PRIMED"] = self.primeVal
                @Property
                @primed_property(primer=prime, initVal=None, unprimeableVal=2)
                @settings_property(name="PRIMED")
                def x(): return {}
                def __init__(self):
                    self.settings={}
                    self.primeVal = "initialized"
            t = Test()
            self.failIf("_PRIMED_prime" in dir(t),
                        getattr(t, "_PRIMED_prime", None))
            self.failUnless(t.x == "initialized", t.x)
            t.x = 1
            self.failUnless(t.x == 1, t.x)
            t.x = None
            self.failUnless(t.x == "initialized", t.x)
            t._PRIMED_prime = True
            t.x = 3
            self.failUnless(t.x == "initialized", t.x)
            t._PRIMED_prime = False
            t.x = 3
            self.failUnless(t.x == 3, t.x)
            # test unprimableVal
            t.x = None
            t.primeVal = None
            self.failUnless(t.x == 2, t.x)
        def testChangeHookLocalProperty(self):
            class Test(object):
                def _hook(self, old, new):
                    self.old = old
                    self.new = new

                @Property
                @change_hook_property(_hook)
                @local_property(name="HOOKED")
                def x(): return {}
            t = Test()
            t.x = 1
            self.failUnless(t.old == None, t.old)
            self.failUnless(t.new == 1, t.new)
            t.x = 1
            self.failUnless(t.old == None, t.old)
            self.failUnless(t.new == 1, t.new)
            t.x = 2
            self.failUnless(t.old == 1, t.old)
            self.failUnless(t.new == 2, t.new)
        def testChangeHookMutableProperty(self):
            class Test(object):
                def _hook(self, old, new):
                    self.old = old
                    self.new = new
                    self.hook_calls += 1

                @Property
                @change_hook_property(_hook, mutable=True)
                @local_property(name="HOOKED")
                def x(): return {}
            t = Test()
            t.hook_calls = 0
            t.x = []
            self.failUnless(t.old == None, t.old)
            self.failUnless(t.new == [], t.new)
            self.failUnless(t.hook_calls == 1, t.hook_calls)
            a = t.x
            a.append(5)
            t.x = a
            self.failUnless(t.old == [], t.old)
            self.failUnless(t.new == [5], t.new)
            self.failUnless(t.hook_calls == 2, t.hook_calls)
            t.x = []
            self.failUnless(t.old == [5], t.old)
            self.failUnless(t.new == [], t.new)
            self.failUnless(t.hook_calls == 3, t.hook_calls)
            # now append without reassigning.  this doesn't trigger the
            # change, since we don't ever set t.x, only get it and mess
            # with it.  It does, however, update our t.new, since t.new =
            # t.x and is not a static copy.
            t.x.append(5)
            self.failUnless(t.old == [5], t.old)
            self.failUnless(t.new == [5], t.new)
            self.failUnless(t.hook_calls == 3, t.hook_calls)
            # however, the next t.x get _will_ notice the change...
            a = t.x
            self.failUnless(t.old == [], t.old)
            self.failUnless(t.new == [5], t.new)
            self.failUnless(t.hook_calls == 4, t.hook_calls)
            t.x.append(6) # this append(6) is not noticed yet
            self.failUnless(t.old == [], t.old)
            self.failUnless(t.new == [5,6], t.new)
            self.failUnless(t.hook_calls == 4, t.hook_calls)
            # this append(7) is not noticed, but the t.x get causes the
            # append(6) to be noticed
            t.x.append(7)
            self.failUnless(t.old == [5], t.old)
            self.failUnless(t.new == [5,6,7], t.new)
            self.failUnless(t.hook_calls == 5, t.hook_calls)
            a = t.x # now the append(7) is noticed
            self.failUnless(t.old == [5,6], t.old)
            self.failUnless(t.new == [5,6,7], t.new)
            self.failUnless(t.hook_calls == 6, t.hook_calls)

    suite = unittest.TestLoader().loadTestsFromTestCase(DecoratorTests)



# Copyright (C) 2008-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Handle ID creation and parsing.

Format
======

BE IDs are formatted::

    <bug-directory>[/<bug>[/<comment>]]

where each ``<..>`` is a UUID.  For example::

    bea86499-824e-4e77-b085-2d581fa9ccab/3438b72c-6244-4f1d-8722-8c8d41484e35

refers to bug ``3438b72c-6244-4f1d-8722-8c8d41484e35`` which is
located in bug directory ``bea86499-824e-4e77-b085-2d581fa9ccab``.
This is a bit of a mouthful, so you can truncate each UUID so long as
it remains unique.  For example::

    bea/343

If there were two bugs ``3438...`` and ``343a...`` in ``bea``, you'd
have to use::

    bea/3438

BE will only truncate each UUID down to three characters to slightly
future-proof the short user ids.  However, if you want to save keystrokes
and you *know* there is only one bug directory, feel free to truncate
all the way to zero characters::

    /3438

Cross references
================

To refer to other bug-directories/bugs/comments from bug comments, simply
enclose the ID in pound signs (``#``).  BE will automatically expand the
truncations to the full UUIDs before storing the comment, and the reference
will be appropriately truncated (and hyperlinked, if possible) when the
comment is displayed.

Scope
=====

Although bug and comment IDs always appear in compound references,
UUIDs at each level are globally unique.  For example, comment
``bea/343/ba96f1c0-ba48-4df8-aaf0-4e3a3144fc46`` will *only* appear
under ``bea/343``.  The prefix (``bea/343``) allows BE to reduce
caching global comment-lookup tables and enables easy error messages
("I couldn't find ``bea/343/ba9`` because I don't know where the
``bea`` bug directory is located").
"""

import os.path
import re

import libbe

if libbe.TESTING == True:
    import doctest
    import sys
    import unittest

try:
    from uuid import uuid4 # Python >= 2.5
    def uuid_gen():
        id = uuid4()
        idstr = id.urn
        start = "urn:uuid:"
        assert idstr.startswith(start)
        return idstr[len(start):]
except ImportError:
    import os
    import sys
    from subprocess import Popen, PIPE

    def uuid_gen():
        # Shell-out to system uuidgen
        args = ['uuidgen', 'r']
        try:
            if sys.platform != "win32":
                q = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            else:
                # win32 don't have os.execvp() so have to run command in a shell
                q = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE,
                          shell=True, cwd=cwd)
        except OSError, e :
            strerror = "%s\nwhile executing %s" % (e.args[1], args)
            raise OSError, strerror
        output, error = q.communicate()
        status = q.wait()
        if status != 0:
            strerror = "%s\nwhile executing %s" % (status, args)
            raise Exception, strerror
        return output.rstrip('\n')


HIERARCHY = ['bugdir', 'bug', 'comment']
"""Keep track of the object type hierarchy.
"""

class MultipleIDMatches (ValueError):
    """Multiple IDs match the given user ID.

    Parameters
    ----------
    id : str
      The not-specific-enough truncated UUID.
    common : str
      The initial characters common to all matching UUIDs.
    matches : list of str
      The list of possibly matching UUIDs.
    """
    def __init__(self, id, common, matches):
        msg = ('More than one id matches %s.  '
               'Please be more specific (%s*).\n%s' % (id, common, matches))
        ValueError.__init__(self, msg)
        self.id = id
        self.common = common
        self.matches = matches

class NoIDMatches (KeyError):
    """No IDs match the given user ID.

    Parameters
    ----------
    id : str
      The not-matching, possibly truncated UUID.
    possible_ids : list of str
      The list of potential UUIDs at that level.
    msg : str, optional
      A helpful message explaining what went wrong.
    """
    def __init__(self, id, possible_ids, msg=None):
        KeyError.__init__(self, id)
        self.id = id
        self.possible_ids = possible_ids
        self.msg = msg
    def __str__(self):
        if self.msg == None:
            return 'No id matches %s.\n%s' % (self.id, self.possible_ids)
        return self.msg

class InvalidIDStructure (KeyError):
    """A purported ID does not have the appropriate syntax.

    Parameters
    ----------
    id : str
      The purported ID.
    msg : str, optional
      A helpful message explaining what went wrong.
    """
    def __init__(self, id, msg=None):
        KeyError.__init__(self, id)
        self.id = id
        self.msg = msg
    def __str__(self):
        if self.msg == None:
            return 'Invalid id structure "%s"' % self.id
        return self.msg

def _assemble(args, check_length=False):
    """Join a bunch of level UUIDs into a single ID.

    See Also
    --------
    _split : inverse
    """
    args = list(args)
    for i,arg in enumerate(args):
        if arg == None:
            args[i] = ''
    id = '/'.join(args)
    if check_length == True:
        assert len(args) > 0, args
        if len(args) > len(HIERARCHY):
            raise InvalidIDStructure(
                id, '%d > %d levels in "%s"' % (len(args), len(HIERARCHY), id))
    return id

def _split(id, check_length=False):
    """Split an ID into a list of level UUIDs.

    See Also
    --------
    _assemble : inverse
    """
    args = id.split('/')
    for i,arg in enumerate(args):
        if arg == '':
            args[i] = None
    if check_length == True:
        assert len(args) > 0, args
        if len(args) > len(HIERARCHY):
            raise InvalidIDStructure(
                id, '%d > %d levels in "%s"' % (len(args), len(HIERARCHY), id))
    return args

def _truncate(uuid, other_uuids, min_length=3):
    """Truncate a UUID to the shortest length >= `min_length` such that it
    is *not* a truncated form of a UUID in `other_uuids`.

    Parameters
    ----------
    uuid : str
      The UUID to truncate.
    other_uuids : list of str
      The other UUIDs which the truncation *might* refer to.  May
      contain `uuid`.
    min_length : int
      Avoid rapidly outdated truncations, even if they are unique now.

    See Also
    --------
    _expand : inverse
    """
    if min_length == -1:
        return uuid
    chars = min_length
    for id in other_uuids:
        if id == uuid:
            continue
        while (id[:chars] == uuid[:chars]):
            chars+=1
    return uuid[:chars]

def _expand(truncated_id, common, other_ids):
    """Expand a truncated UUID.

    Parameters
    ----------
    truncated_id : str
      The ID to expand.
    common : str
      The common portion `truncated_id` shares with the UUIDs in
      `other_ids`.  Not used by ``_expand``, but passed on to the
      matching exceptions if they occur.
    other_uuids : list of str
      The other UUIDs which the truncation *might* refer to.  May
      contain `uuid`.

    Raises
    ------
    NoIDMatches
    MultipleIDMatches

    See Also
    --------
    _expand : inverse
    """
    other_ids = list(other_ids)
    if len(other_ids) == 0:
        raise NoIDMatches(truncated_id, other_ids)
    if truncated_id == None:
        if len(other_ids) == 1:
            return other_ids[0]
        raise MultipleIDMatches(truncated_id, common, other_ids)
    matches = []
    other_ids = list(other_ids)
    for id in other_ids:
        if id.startswith(truncated_id):
            if id == truncated_id:
                return id
            matches.append(id)
    if len(matches) > 1:
        raise MultipleIDMatches(truncated_id, common, matches)
    if len(matches) == 0:
        raise NoIDMatches(truncated_id, other_ids)
    return matches[0]


class ID (object):
    """Store an object ID and produce various representations.

    Parameters
    ----------
    object : :py:class:`~libbe.bugdir.BugDir` or :py:class:`~libbe.bug.Bug` or :py:class:`~libbe.comment.Comment`
      The object that the ID applies to.
    type : 'bugdir' or 'bug' or 'comment'
      The type of the object.

    Notes
    -----

    IDs have several formats specialized for different uses.

    In storage, all objects are represented by their uuid alone,
    because that is the simplest globally unique identifier.  You can
    generate ids of this sort with the .storage() method.  Because an
    object's storage may be distributed across several chunks, and the
    chunks may not have their own uuid, we generate chunk ids by
    prepending the objects uuid to the chunk name.  The user id types
    do not support this chunk extension feature.

    For users, the full uuids are a bit overwhelming, so we truncate
    them while retaining local uniqueness (with regards to the other
    objects currently in storage).  We also prepend truncated parent
    ids for two reasons:

    1. So that a user can locate the repository containing the
       referenced object.  It would be hard to find bug ``XYZ`` if
       that's all you knew.  Much easier with ``ABC/XYZ``, where
       ``ABC`` is the bugdir.  Each project can publish a list of
       bugdir-id-to-location mappings, e.g.::

            ABC...(full uuid)...DEF   https://server.com/projectX/be/

       which is easier than publishing all-object-ids-to-location
       mappings.

    2. Because it's easier to generate and parse truncated ids if you
       don't have to fetch all the ids in the storage repository but
       can restrict yourself to a specific branch.

    You can generate ids of this sort with the :py:meth:`user` method,
    although in order to preform the truncation, your object (and its
    parents must define a `sibling_uuids` method.

    While users can use the convenient short user ids in the short
    term, the truncation will inevitably lead to name collision.  To
    avoid that, we provide a non-truncated form of the short user ids
    via the :py:meth:`long_user` method.  These long user ids should be
    converted to short user ids by intelligent user interfaces.

    See Also
    --------
    parse_user : get uuids back out of the user ids.
    short_to_long_user : convert a single short user id to a long user id.
    long_to_short_user : convert a single long user id to a short user id.
    short_to_long_text : scan text for user ids & convert to long user ids.
    long_to_short_text : scan text for long user ids & convert to short user ids.
    """
    def __init__(self, object, type):
        self._object = object
        self._type = type
        assert self._type in HIERARCHY, self._type

    def storage(self, *args):
        return _assemble([self._object.uuid]+list(args))

    def _ancestors(self):
        ret = [self._object]
        index = HIERARCHY.index(self._type)
        if index == 0:
            return ret
        o = self._object
        for i in range(index, 0, -1):
            parent_name = HIERARCHY[i-1]
            o = getattr(o, parent_name, None)
            ret.insert(0, o)
        return ret

    def long_user(self):
        return _assemble([o.uuid for o in self._ancestors()],
                         check_length=True)

    def user(self):
        ids = []
        for o in self._ancestors():
            if o == None:
                ids.append(None)
            else:
                ids.append(_truncate(o.uuid, o.sibling_uuids()))
        return _assemble(ids, check_length=True)

def child_uuids(child_storage_ids):
    """Extract uuid children from other children generated by
    :py:meth:`ID.storage`.

    This is useful for separating data belonging to a particular
    object directly from entries for its child objects.  Since the
    :py:class:`~libbe.storage.base.Storage` backend doesn't distinguish
    between the two.

    Examples
    --------

    >>> list(child_uuids(['abc123/values', '123abc', '123def']))
    ['123abc', '123def']
    """
    for id in child_storage_ids:
        fields = _split(id)
        if len(fields) == 1:
            yield fields[0]

def long_to_short_user(bugdirs, id):
    """Convert a long user ID to a short user ID (see :py:class:`ID`).
    The list of bugdirs allows uniqueness-maintaining truncation of
    the bugdir portion of the ID.

    See Also
    --------
    short_to_long_user : inverse
    long_to_short_text : conversion on a block of text
    """
    ids = _split(id, check_length=True)
    matching_bugdirs = [bd for bd in bugdirs.values() if bd.uuid == ids[0]]
    if len(matching_bugdirs) == 0:
        raise NoIDMatches(id, [bd.uuid for bd in bugdirs.values()])
    elif len(matching_bugdirs) > 1:
        raise MultipleIDMatches(id, '', [bd.uuid for bd in bugdirs.values()])
    bugdir = matching_bugdirs[0]
    objects = [bugdir]
    if len(ids) >= 2:
        bug = bugdir.bug_from_uuid(ids[1])
        objects.append(bug)
    if len(ids) >= 3:
        comment = bug.comment_from_uuid(ids[2])
        objects.append(comment)
    for i,obj in enumerate(objects):
        ids[i] = _truncate(ids[i], obj.sibling_uuids())
    return _assemble(ids)

def short_to_long_user(bugdirs, id):
    """Convert a short user ID to a long user ID (see :py:class:`ID`).  The
    list of bugdirs allows uniqueness-checking during expansion of the
    bugdir portion of the ID.

    See Also
    --------
    long_to_short_user : inverse
    short_to_long_text : conversion on a block of text
    """
    ids = _split(id, check_length=True)
    ids[0] = _expand(ids[0], common=None,
                     other_ids=[bd.uuid for bd in bugdirs.values()])
    if len(ids) == 1:
        return _assemble(ids)
    bugdir = [bd for bd in bugdirs.values() if bd.uuid == ids[0]][0]
    ids[1] = _expand(ids[1], common=bugdir.id.user(),
                     other_ids=bugdir.uuids())
    if len(ids) == 2:
        return _assemble(ids)
    bug = bugdir.bug_from_uuid(ids[1])
    ids[2] = _expand(ids[2], common=bug.id.user(),
                     other_ids=bug.uuids())
    return _assemble(ids)


REGEXP = '#([-a-f0-9]*)(/[-a-g0-9]*)?(/[-a-g0-9]*)?#'
"""Regular expression for matching IDs (both short and long) in text.
"""

class IDreplacer (object):
    """Helper class for ID replacement in text.

    Reassembles the match elements from :py:data:`REGEXP` matching
    into the original ID, for easier replacement.

    See Also
    --------
    short_to_long_text, long_to_short_text
    """
    def __init__(self, bugdirs, replace_fn, wrap=True, strict=True):
        self.bugdirs = bugdirs
        self.replace_fn = replace_fn
        self.wrap = wrap
        self.strict = strict

    def __call__(self, match):
        ids = []
        for m in match.groups():
            if m == None:
                m = ''
            ids.append(m)
        try:
            replacement = self.replace_fn(self.bugdirs, ''.join(ids))
        except (MultipleIDMatches, NoIDMatches, InvalidIDStructure):
            if self.strict:
                raise
            replacement = ''.join(ids)
        if self.wrap == True:
            return '#%s#' % replacement
        return replacement

def short_to_long_text(bugdirs, text):
    """Convert short user IDs to long user IDs in text (see :py:class:`ID`).
    The list of bugdirs allows uniqueness-checking during expansion of
    the bugdir portion of the ID.

    See Also
    --------
    short_to_long_user : conversion on a single ID
    long_to_short_text : inverse
    """
    return re.sub(
        REGEXP, IDreplacer(bugdirs, short_to_long_user, strict=False), text)

def long_to_short_text(bugdirs, text):
    """Convert long user IDs to short user IDs in text (see :py:class:`ID`).
    The list of bugdirs allows uniqueness-maintaining truncation of
    the bugdir portion of the ID.

    See Also
    --------
    long_to_short_user : conversion on a single ID
    short_to_long_text : inverse
    """
    return re.sub(REGEXP, IDreplacer(bugdirs, long_to_short_user), text)

def residual(base, fragment):
    """Split the short ID `fragment` into a portion corresponding
    to `base`, and a portion inside `base`.

    Examples
    --------

    >>> residual('ABC/DEF/', '//GHI')
    ('//', 'GHI')
    >>> residual('ABC/DEF/', '/D/GHI')
    ('/D/', 'GHI')
    >>> residual('ABC/DEF', 'A/D/GHI')
    ('A/D/', 'GHI')
    >>> residual('ABC/DEF', 'A/D/GHI/JKL')
    ('A/D/', 'GHI/JKL')
    """
    base = base.rstrip('/') + '/'
    ids = fragment.split('/')
    base_count = base.count('/')
    root_ids = ids[:base_count] + ['']
    residual_ids = ids[base_count:]
    return ('/'.join(root_ids), '/'.join(residual_ids))

def _parse_user(id):
    """Parse a user ID (see :py:class:`ID`), returning a dict of parsed
    information.

    The returned dict will contain a value for "type" (from
    :py:data:`HIERARCHY`) and values for the levels that are defined.

    Examples
    --------

    >>> _parse_user('ABC/DEF/GHI') == \\
    ...     {'bugdir':'ABC', 'bug':'DEF', 'comment':'GHI', 'type':'comment'}
    True
    >>> _parse_user('ABC/DEF') == \\
    ...     {'bugdir':'ABC', 'bug':'DEF', 'type':'bug'}
    True
    >>> _parse_user('ABC') == \\
    ...     {'bugdir':'ABC', 'type':'bugdir'}
    True
    >>> _parse_user('') == \\
    ...     {'bugdir':None, 'type':'bugdir'}
    True
    >>> _parse_user('/') == \\
    ...     {'bugdir':None, 'bug':None, 'type':'bug'}
    True
    >>> _parse_user('/DEF/') == \\
    ...     {'bugdir':None, 'bug':'DEF', 'comment':None, 'type':'comment'}
    True
    >>> _parse_user('a/b/c/d')
    Traceback (most recent call last): 
      ...
    InvalidIDStructure: 4 > 3 levels in "a/b/c/d"
    """
    ret = {}
    args = _split(id, check_length=True)
    for i,(type,arg) in enumerate(zip(HIERARCHY, args)):
        if arg != None and len(arg) == 0:
            raise InvalidIDStructure(
                id, 'Invalid %s part %d "%s" of id "%s"' % (type, i, arg, id))
        ret['type'] = type
        ret[type] = arg
    return ret

def parse_user(bugdirs, id):
    """Parse a user ID (see :py:class:`ID`), returning a dict of parsed
    information.

    The returned dict will contain a value for "type" (from
    :py:data:`HIERARCHY`) and values for the levels that are defined.

    Notes
    -----
    This function tries to expand IDs before parsing, so it can handle
    both short and long IDs successfully.
    """
    long_id = short_to_long_user(bugdirs, id)
    return _parse_user(long_id)

if libbe.TESTING == True:
    class UUIDtestCase(unittest.TestCase):
        def testUUID_gen(self):
            id = uuid_gen()
            self.failUnless(len(id) == 36, 'invalid UUID "%s"' % id)

    class DummyObject (object):
        def __init__(self, uuid, parent=None, siblings=[]):
            self.uuid = uuid
            self._siblings = siblings
            if parent == None:
                type_i = 0
            else:
                assert parent.type in HIERARCHY, parent
                setattr(self, parent.type, parent)
                type_i = HIERARCHY.index(parent.type) + 1
            self.type = HIERARCHY[type_i]
            self.id = ID(self, self.type)
        def sibling_uuids(self):
            return self._siblings

    class IDtestCase(unittest.TestCase):
        def setUp(self):
            self.bugdir = DummyObject('1234abcd')
            self.bug = DummyObject('abcdef', self.bugdir, ['a1234', 'ab9876'])
            self.comment = DummyObject('12345678', self.bug, ['1234abcd', '1234cdef'])
            self.bd_id = self.bugdir.id
            self.b_id = self.bug.id
            self.c_id = self.comment.id
        def test_storage(self):
            self.failUnless(self.bd_id.storage() == self.bugdir.uuid,
                            self.bd_id.storage())
            self.failUnless(self.b_id.storage() == self.bug.uuid,
                            self.b_id.storage())
            self.failUnless(self.c_id.storage() == self.comment.uuid,
                            self.c_id.storage())
            self.failUnless(self.bd_id.storage('x', 'y', 'z') == \
                                '1234abcd/x/y/z',
                            self.bd_id.storage('x', 'y', 'z'))
        def test_long_user(self):
            self.failUnless(self.bd_id.long_user() == self.bugdir.uuid,
                            self.bd_id.long_user())
            self.failUnless(self.b_id.long_user() == \
                                '/'.join([self.bugdir.uuid, self.bug.uuid]),
                            self.b_id.long_user())
            self.failUnless(self.c_id.long_user() ==
                                '/'.join([self.bugdir.uuid, self.bug.uuid,
                                          self.comment.uuid]),
                            self.c_id.long_user)
        def test_user(self):
            self.failUnless(self.bd_id.user() == '123',
                            self.bd_id.user())
            self.failUnless(self.b_id.user() == '123/abc',
                            self.b_id.user())
            self.failUnless(self.c_id.user() == '123/abc/12345',
                            self.c_id.user())

    class ShortLongParseTestCase(unittest.TestCase):
        def setUp(self):
            self.bugdir = DummyObject('1234abcd')
            self.bugdirs = {self.bugdir.uuid: self.bugdir}
            self.bug = DummyObject('abcdef', self.bugdir, ['a1234', 'ab9876'])
            self.comment = DummyObject('12345678', self.bug, ['1234abcd', '1234cdef'])
            self.bd_id = self.bugdir.id
            self.b_id = self.bug.id
            self.c_id = self.comment.id
            self.bugdir.bug_from_uuid = lambda uuid: self.bug
            self.bugdir.uuids = lambda : self.bug.sibling_uuids() + [self.bug.uuid]
            self.bug.comment_from_uuid = lambda uuid: self.comment
            self.bug.uuids = lambda : self.comment.sibling_uuids() + [self.comment.uuid]
            self.short = 'bla bla #123/abc# bla bla #123/abc/12345# bla bla'
            self.long = 'bla bla #1234abcd/abcdef# bla bla #1234abcd/abcdef/12345678# bla bla'
            self.short_id_parse_pairs = [
                ('', {'bugdir':'1234abcd', 'type':'bugdir'}),
                ('123/abc', {'bugdir':'1234abcd', 'bug':'abcdef',
                             'type':'bug'}),
                ('123/abc/12345', {'bugdir':'1234abcd', 'bug':'abcdef',
                                   'comment':'12345678', 'type':'comment'}),
                ]
            self.short_id_exception_pairs = [
                ('z', NoIDMatches('z', ['1234abcd'])),
                ('///', InvalidIDStructure(
                        '///', msg='4 > 3 levels in "///"')),
                ('/', MultipleIDMatches(
                        None, '123', ['a1234', 'ab9876', 'abcdef'])),
                ('123/', MultipleIDMatches(
                        None, '123', ['a1234', 'ab9876', 'abcdef'])),
                ('123/abc/', MultipleIDMatches(
                        None, '123/abc', ['1234abcd','1234cdef','12345678'])),
                ]
        def test_short_to_long_text(self):
            self.failUnless(short_to_long_text(
                    self.bugdirs, self.short) == self.long,
                            '\n' + self.short + '\n' + short_to_long_text(
                    self.bugdirs, self.short) + '\n' + self.long)
        def test_long_to_short_text(self):
            self.failUnless(long_to_short_text(
                    self.bugdirs, self.long) == self.short,
                            '\n' + long_to_short_text(
                    self.bugdirs, self.long
                    ) + '\n' + self.short)
        def test_parse_user(self):
            for short_id,parsed in self.short_id_parse_pairs:
                ret = parse_user(self.bugdirs, short_id)
                self.failUnless(ret == parsed,
                                'got %s\nexpected %s' % (ret, parsed))
        def test_parse_user_exceptions(self):
            for short_id,exception in self.short_id_exception_pairs:
                try:
                    ret = parse_user(self.bugdirs, short_id)
                    self.fail('Expected parse_user(bugdir, "%s") to raise %s,'
                              '\n  but it returned %s'
                              % (short_id, exception.__class__.__name__, ret))
                except exception.__class__, e:
                    for attr in dir(e):
                        if attr.startswith('_') or attr == 'args':
                            continue
                        value = getattr(e, attr)
                        expected = getattr(exception, attr)
                        self.failUnless(
                            value == expected,
                            'Expected parse_user(bugdir, "%s") %s.%s'
                            '\n  to be %s, but it is %s\n\n%s'
                              % (short_id, exception.__class__.__name__,
                                 attr, expected, value, e))

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define the :py:class:`~libbe.storage.base.Storage` and
:py:class:`~libbe.storage.base.VersionedStorage` classes for storing BE
data.

Also define assorted implementations for the Storage classes:

* :py:mod:`libbe.storage.vcs`
* :py:mod:`libbe.storage.http`

Also define an assortment of storage-related tools and utilities:

* :py:mod:`libbe.storage.util`
"""

from . import base as _base

ConnectionError = _base.ConnectionError
InvalidStorageVersion = _base.InvalidStorageVersion
InvalidID = _base.InvalidID
InvalidRevision = _base.InvalidRevision
InvalidDirectory = _base.InvalidDirectory
NotWriteable = _base.NotWriteable
NotReadable = _base.NotReadable
EmptyCommit = _base.EmptyCommit

# a list of all past versions
STORAGE_VERSIONS = ['Bugs Everywhere Tree 1 0',
                    'Bugs Everywhere Directory v1.1',
                    'Bugs Everywhere Directory v1.2',
                    'Bugs Everywhere Directory v1.3',
                    'Bugs Everywhere Directory v1.4',
                    'Bugs Everywhere Directory v1.5',
                    ]

# the current version
STORAGE_VERSION = STORAGE_VERSIONS[-1]

def get_http_storage(location):
    import http
    return http.HTTP(location)

def get_vcs_storage(location):
    import vcs
    s = vcs.detect_vcs(location)
    s.repo = location
    return s

def get_storage(location):
    """
    Return a Storage instance from a repo location string.
    """
    if location.startswith('http://') or location.startswith('https://'):
        return get_http_storage(location)
    return get_vcs_storage(location)

__all__ = [ConnectionError, InvalidStorageVersion, InvalidID,
           InvalidRevision, InvalidDirectory, NotWriteable, NotReadable,
           EmptyCommit, STORAGE_VERSIONS, STORAGE_VERSION,
           get_storage]



# Copyright (C) 2008-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Niall Douglas (s_sourceforge@nedprod.com) <spam@spamtrap.com>
#                         Robert Lehmann <mail@robertlehmann.de>
#                         Thomas Habets <thomas@habets.pp.se>
#                         Valtteri Kokkoniemi <rvk@iki.fi>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define :py:class:`Bug` for representing bugs.
"""

import copy
import os
import os.path
import errno
import sys
import time
import types
try: # import core module, Python >= 2.5
    from xml.etree import ElementTree
except ImportError: # look for non-core module
    from elementtree import ElementTree
import xml.sax.saxutils

import libbe
import libbe.util.id
from libbe.storage.util.properties import Property, doc_property, \
    local_property, defaulting_property, checked_property, cached_property, \
    primed_property, change_hook_property, settings_property
import libbe.storage.util.settings_object as settings_object
import libbe.storage.util.mapfile as mapfile
import libbe.comment as comment
import libbe.util.utility as utility

if libbe.TESTING == True:
    import doctest


### Define and describe valid bug categories
# Use a tuple of (category, description) tuples since we don't have
# ordered dicts in Python yet http://www.python.org/dev/peps/pep-0372/

# in order of increasing severity.  (name, description) pairs
severity_def = (
  ("target", "The issue is a target or milestone, not a bug."),
  ("wishlist","A feature that could improve usefulness, but not a bug."),
  ("minor","The standard bug level."),
  ("serious","A bug that requires workarounds."),
  ("critical","A bug that prevents some features from working at all."),
  ("fatal","A bug that makes the package unusable."))

# in order of increasing resolution
# roughly following http://www.bugzilla.org/docs/3.2/en/html/lifecycle.html
active_status_def = (
  ("unconfirmed","A possible bug which lacks independent existance confirmation."),
  ("open","A working bug that has not been assigned to a developer."),
  ("assigned","A working bug that has been assigned to a developer."),
  ("test","The code has been adjusted, but the fix is still being tested."))
inactive_status_def = (
  ("closed", "The bug is no longer relevant."),
  ("fixed", "The bug should no longer occur."),
  ("wontfix","It's not a bug, it's a feature."))


### Convert the description tuples to more useful formats

severity_values = ()
severity_description = {}
severity_index = {}
def load_severities(severity_def):
    global severity_values
    global severity_description
    global severity_index
    if severity_def == None:
        return
    severity_values = tuple([val for val,description in severity_def])
    severity_description = dict(severity_def)
    severity_index = {}
    for i,severity in enumerate(severity_values):
        severity_index[severity] = i
load_severities(severity_def)

active_status_values = []
inactive_status_values = []
status_values = []
status_description = {}
status_index = {}
def load_status(active_status_def, inactive_status_def):
    global active_status_values
    global inactive_status_values
    global status_values
    global status_description
    global status_index
    if active_status_def == None:
        active_status_def = globals()["active_status_def"]
    if inactive_status_def == None:
        inactive_status_def = globals()["inactive_status_def"]
    active_status_values = tuple([val for val,description in active_status_def])
    inactive_status_values = tuple([val for val,description in inactive_status_def])
    status_values = active_status_values + inactive_status_values
    status_description = dict(tuple(active_status_def) + tuple(inactive_status_def))
    status_index = {}
    for i,status in enumerate(status_values):
        status_index[status] = i
load_status(active_status_def, inactive_status_def)


class Bug (settings_object.SavedSettingsObject):
    """A bug (or issue) is a place to store attributes and attach
    :py:class:`~libbe.comment.Comment`\s.  In mailing-list terms, a bug is
    analogous to a thread.  Bugs are normally stored in
    :py:class:`~libbe.bugdir.BugDir`\s.

    >>> b = Bug()
    >>> print b.status
    open
    >>> print b.severity
    minor

    There are two formats for time, int and string.  Setting either
    one will adjust the other appropriately.  The string form is the
    one stored in the bug's settings file on disk.

    >>> print type(b.time)
    <type 'int'>
    >>> print type(b.time_string)
    <type 'str'>
    >>> b.time = 0
    >>> print b.time_string
    Thu, 01 Jan 1970 00:00:00 +0000
    >>> b.time_string="Thu, 01 Jan 1970 00:01:00 +0000"
    >>> b.time
    60
    >>> print b.settings["time"]
    Thu, 01 Jan 1970 00:01:00 +0000
    """
    settings_properties = []
    required_saved_properties = []
    _prop_save_settings = settings_object.prop_save_settings
    _prop_load_settings = settings_object.prop_load_settings
    def _versioned_property(settings_properties=settings_properties,
                            required_saved_properties=required_saved_properties,
                            **kwargs):
        if "settings_properties" not in kwargs:
            kwargs["settings_properties"] = settings_properties
        if "required_saved_properties" not in kwargs:
            kwargs["required_saved_properties"]=required_saved_properties
        return settings_object.versioned_property(**kwargs)

    @_versioned_property(name="severity",
                         doc="A measure of the bug's importance",
                         default="minor",
                         check_fn=lambda s: s in severity_values,
                         require_save=True)
    def severity(): return {}

    @_versioned_property(name="status",
                         doc="The bug's current status",
                         default="open",
                         check_fn=lambda s: s in status_values,
                         require_save=True)
    def status(): return {}

    @property
    def active(self):
        return self.status in active_status_values

    @_versioned_property(name="creator",
                         doc="The user who entered the bug into the system")
    def creator(): return {}

    @_versioned_property(name="reporter",
                         doc="The user who reported the bug")
    def reporter(): return {}

    @_versioned_property(name="assigned",
                         doc="The developer in charge of the bug")
    def assigned(): return {}

    @_versioned_property(name="time",
                         doc="An RFC 2822 timestamp for bug creation")
    def time_string(): return {}

    def _get_time(self):
        if self.time_string == None:
            self._cached_time_string = None
            self._cached_time = None
            return None
        if (not hasattr(self, '_cached_time_string')
            or self.time_string != self._cached_time_string):
            self._cached_time_string = self.time_string
            self._cached_time = utility.str_to_time(self.time_string)
        return self._cached_time
    def _set_time(self, value):
        if not hasattr(self, '_cached_time') or value != self._cached_time:
            self.time_string = utility.time_to_str(value)
            self._cached_time_string = self.time_string
            self._cached_time = value
    time = property(fget=_get_time,
                    fset=_set_time,
                    doc="An integer version of .time_string")

    def _extra_strings_check_fn(value):
        return utility.iterable_full_of_strings(value, \
                         alternative=settings_object.EMPTY)
    def _extra_strings_change_hook(self, old, new):
        self.extra_strings.sort() # to make merging easier
        self._prop_save_settings(old, new)
    @_versioned_property(name="extra_strings",
                         doc="Space for an array of extra strings.  Useful for storing state for functionality implemented purely in becommands/<some_function>.py.",
                         default=[],
                         check_fn=_extra_strings_check_fn,
                         change_hook=_extra_strings_change_hook,
                         mutable=True)
    def extra_strings(): return {}

    @_versioned_property(name="summary",
                         doc="A one-line bug description")
    def summary(): return {}

    def _get_comment_root(self, load_full=False):
        if self.storage != None and self.storage.is_readable():
            return comment.load_comments(self, load_full=load_full)
        else:
            return comment.Comment(self, uuid=comment.INVALID_UUID)

    @Property
    @cached_property(generator=_get_comment_root)
    @local_property("comment_root")
    @doc_property(doc="The trunk of the comment tree.  We use a dummy root comment by default, because there can be several comment threads rooted on the same parent bug.  To simplify comment interaction, we condense these threads into a single thread with a Comment dummy root.")
    def comment_root(): return {}

    def __init__(self, bugdir=None, uuid=None, from_storage=False,
                 load_comments=False, summary=None):
        settings_object.SavedSettingsObject.__init__(self)
        self.bugdir = bugdir
        self.storage = None
        self.uuid = uuid
        self.id = libbe.util.id.ID(self, 'bug')
        if from_storage == False:
            if uuid == None:
                self.uuid = libbe.util.id.uuid_gen()
            self.time = int(time.time()) # only save to second precision
            self.summary = summary
            dummy = self.comment_root
        if self.bugdir != None:
            self.storage = self.bugdir.storage
        if from_storage == False:
            if self.storage != None and self.storage.is_writeable():
                self.save()

    def __repr__(self):
        return "Bug(uuid=%r)" % self.uuid

    def __str__(self):
        return self.string(shortlist=True)

    def __cmp__(self, other):
        return cmp_full(self, other)

    # serializing methods

    def _setting_attr_string(self, setting):
        value = getattr(self, setting)
        if value == None:
            return ""
        if type(value) not in types.StringTypes:
            return str(value)
        return value

    def string(self, shortlist=False, show_comments=False):
        if shortlist == False:
            if self.time == None:
                timestring = ""
            else:
                htime = utility.handy_time(self.time)
                timestring = "%s (%s)" % (htime, self.time_string)
            info = [("ID", self.uuid),
                    ("Short name", self.id.user()),
                    ("Severity", self.severity),
                    ("Status", self.status),
                    ("Assigned", self._setting_attr_string("assigned")),
                    ("Reporter", self._setting_attr_string("reporter")),
                    ("Creator", self._setting_attr_string("creator")),
                    ("Created", timestring)]
            for estr in self.extra_strings:
                info.append(('Extra string', estr))
            longest_key_len = max([len(k) for k,v in info])
            infolines = ["  %*s : %s\n" %(longest_key_len,k,v) for k,v in info]
            bugout = "".join(infolines) + "%s" % self.summary.rstrip('\n')
        else:
            statuschar = self.status[0]
            severitychar = self.severity[0]
            chars = "%c%c" % (statuschar, severitychar)
            bugout = "%s:%s: %s" % (self.id.user(),chars,self.summary.rstrip('\n'))

        if show_comments == True:
            self.comment_root.sort(cmp=libbe.comment.cmp_time, reverse=True)
            comout = self.comment_root.string_thread(flatten=False)
            output = bugout + '\n' + comout.rstrip('\n')
        else :
            output = bugout
        return output

    def xml(self, indent=0, show_comments=False):
        """
        >>> bugA = Bug(uuid='0123', summary='Need to test Bug.xml()')
        >>> bugA.uuid = 'bugA'
        >>> bugA.time_string = 'Thu, 01 Jan 1970 00:00:00 +0000'
        >>> bugA.creator = u'Frank'
        >>> bugA.extra_strings += ['TAG: very helpful']
        >>> commA = bugA.comment_root.new_reply(body='comment A')
        >>> commA.uuid = 'commA'
        >>> commA.date = 'Thu, 01 Jan 1970 00:01:00 +0000'
        >>> commB = commA.new_reply(body='comment B')
        >>> commB.uuid = 'commB'
        >>> commB.date = 'Thu, 01 Jan 1970 00:02:00 +0000'
        >>> commC = commB.new_reply(body='comment C')
        >>> commC.uuid = 'commC'
        >>> commC.date = 'Thu, 01 Jan 1970 00:03:00 +0000'
        >>> print(bugA.xml(show_comments=True))  # doctest: +REPORT_UDIFF
        <bug>
          <uuid>bugA</uuid>
          <short-name>/bug</short-name>
          <severity>minor</severity>
          <status>open</status>
          <creator>Frank</creator>
          <created>Thu, 01 Jan 1970 00:00:00 +0000</created>
          <summary>Need to test Bug.xml()</summary>
          <extra-string>TAG: very helpful</extra-string>
          <comment>
            <uuid>commA</uuid>
            <short-name>/bug/commA</short-name>
            <author></author>
            <date>Thu, 01 Jan 1970 00:01:00 +0000</date>
            <content-type>text/plain</content-type>
            <body>comment A</body>
          </comment>
          <comment>
            <uuid>commB</uuid>
            <short-name>/bug/commB</short-name>
            <in-reply-to>commA</in-reply-to>
            <author></author>
            <date>Thu, 01 Jan 1970 00:02:00 +0000</date>
            <content-type>text/plain</content-type>
            <body>comment B</body>
          </comment>
          <comment>
            <uuid>commC</uuid>
            <short-name>/bug/commC</short-name>
            <in-reply-to>commB</in-reply-to>
            <author></author>
            <date>Thu, 01 Jan 1970 00:03:00 +0000</date>
            <content-type>text/plain</content-type>
            <body>comment C</body>
          </comment>
        </bug>
        >>> print(bugA.xml(show_comments=True, indent=2))
        ... # doctest: +REPORT_UDIFF
          <bug>
            <uuid>bugA</uuid>
            <short-name>/bug</short-name>
            <severity>minor</severity>
            <status>open</status>
            <creator>Frank</creator>
            <created>Thu, 01 Jan 1970 00:00:00 +0000</created>
            <summary>Need to test Bug.xml()</summary>
            <extra-string>TAG: very helpful</extra-string>
            <comment>
              <uuid>commA</uuid>
              <short-name>/bug/commA</short-name>
              <author></author>
              <date>Thu, 01 Jan 1970 00:01:00 +0000</date>
              <content-type>text/plain</content-type>
              <body>comment A</body>
            </comment>
            <comment>
              <uuid>commB</uuid>
              <short-name>/bug/commB</short-name>
              <in-reply-to>commA</in-reply-to>
              <author></author>
              <date>Thu, 01 Jan 1970 00:02:00 +0000</date>
              <content-type>text/plain</content-type>
              <body>comment B</body>
            </comment>
            <comment>
              <uuid>commC</uuid>
              <short-name>/bug/commC</short-name>
              <in-reply-to>commB</in-reply-to>
              <author></author>
              <date>Thu, 01 Jan 1970 00:03:00 +0000</date>
              <content-type>text/plain</content-type>
              <body>comment C</body>
            </comment>
          </bug>
        """
        if self.time == None:
            timestring = ""
        else:
            timestring = utility.time_to_str(self.time)

        info = [('uuid', self.uuid),
                ('short-name', self.id.user()),
                ('severity', self.severity),
                ('status', self.status),
                ('assigned', self.assigned),
                ('reporter', self.reporter),
                ('creator', self.creator),
                ('created', timestring),
                ('summary', self.summary)]
        lines = ['<bug>']
        for (k,v) in info:
            if v is not None:
                lines.append('  <%s>%s</%s>' % (k,xml.sax.saxutils.escape(v),k))
        for estr in self.extra_strings:
            lines.append('  <extra-string>%s</extra-string>' % estr)
        if show_comments == True:
            comout = self.comment_root.xml_thread(indent=indent+2)
            if comout:
                comout = comout[indent:]  # strip leading indent spaces
                lines.append(comout)
        lines.append('</bug>')
        istring = ' '*indent
        sep = '\n' + istring
        return istring + sep.join(lines).rstrip('\n')

    def from_xml(self, xml_string, preserve_uuids=False, verbose=True):
        u"""
        Note: If a bug uuid is given, set .alt_id to it's value.
        >>> bugA = Bug(uuid="0123", summary="Need to test Bug.from_xml()")
        >>> bugA.date = "Thu, 01 Jan 1970 00:00:00 +0000"
        >>> bugA.creator = u'Fran\xe7ois'
        >>> bugA.extra_strings += ['TAG: very helpful']
        >>> commA = bugA.comment_root.new_reply(body='comment A')
        >>> commB = bugA.comment_root.new_reply(body='comment B')
        >>> commC = commA.new_reply(body='comment C')
        >>> xml = bugA.xml(show_comments=True)
        >>> bugB = Bug()
        >>> bugB.from_xml(xml, verbose=True)
        >>> bugB.xml(show_comments=True) == xml
        False
        >>> bugB.uuid = bugB.alt_id
        >>> for comm in bugB.comments():
        ...     comm.uuid = comm.alt_id
        ...     comm.alt_id = None
        >>> bugB.xml(show_comments=True) == xml
        True
        >>> bugB.explicit_attrs  # doctest: +NORMALIZE_WHITESPACE
        ['severity', 'status', 'creator', 'time', 'summary']
        >>> len(list(bugB.comments()))
        3
        >>> bugC = Bug()
        >>> bugC.from_xml(xml, preserve_uuids=True)
        >>> bugC.uuid == bugA.uuid
        True
        """
        if type(xml_string) == types.UnicodeType:
            xml_string = xml_string.strip().encode('unicode_escape')
        if hasattr(xml_string, 'getchildren'): # already an ElementTree Element
            bug = xml_string
        else:
            bug = ElementTree.XML(xml_string)
        if bug.tag != 'bug':
            raise utility.InvalidXML( \
                'bug', bug, 'root element must be <bug>')
        tags=['uuid','short-name','severity','status','assigned',
              'reporter', 'creator','created','summary','extra-string']
        self.explicit_attrs = []
        uuid = None
        estrs = []
        comments = []
        for child in bug.getchildren():
            if child.tag == 'short-name':
                pass
            elif child.tag == 'comment':
                comm = comment.Comment(bug=self)
                comm.from_xml(
                    child, preserve_uuids=preserve_uuids, verbose=verbose)
                comments.append(comm)
                continue
            elif child.tag in tags:
                if child.text == None or len(child.text) == 0:
                    text = settings_object.EMPTY
                else:
                    text = xml.sax.saxutils.unescape(child.text)
                    if not isinstance(text, unicode):
                        text = text.decode('unicode_escape')
                    text = text.strip()
                if child.tag == 'uuid' and not preserve_uuids:
                    uuid = text
                    continue # don't set the bug's uuid tag.
                elif child.tag == 'created':
                    if text is not settings_object.EMPTY:
                        self.time = utility.str_to_time(text)
                        self.explicit_attrs.append('time')
                    continue
                elif child.tag == 'extra-string':
                    estrs.append(text)
                    continue # don't set the bug's extra_string yet.
                attr_name = child.tag.replace('-','_')
                self.explicit_attrs.append(attr_name)
                setattr(self, attr_name, text)
            elif verbose == True:
                print >> sys.stderr, 'Ignoring unknown tag %s in %s' \
                    % (child.tag, comment.tag)
        if uuid != self.uuid:
            if not hasattr(self, 'alt_id') or self.alt_id == None:
                self.alt_id = uuid
        self.extra_strings = estrs
        self.add_comments(comments, ignore_missing_references=True)

    def add_comment(self, comment, *args, **kwargs):
        """
        Add a comment too the current bug, under the parent specified
        by comment.in_reply_to.
        Note: If a bug uuid is given, set .alt_id to it's value.

        >>> bugA = Bug(uuid='0123', summary='Need to test Bug.add_comment()')
        >>> bugA.creator = 'Jack'
        >>> commA = bugA.comment_root.new_reply(body='comment A')
        >>> commA.uuid = 'commA'
        >>> commB = comment.Comment(body='comment B')
        >>> commB.uuid = 'commB'
        >>> bugA.add_comment(commB)
        >>> commC = comment.Comment(body='comment C')
        >>> commC.uuid = 'commC'
        >>> commC.in_reply_to = commA.uuid
        >>> bugA.add_comment(commC)
        >>> print bugA.xml(show_comments=True)  # doctest: +ELLIPSIS
        <bug>
          <uuid>0123</uuid>
          <short-name>/012</short-name>
          <severity>minor</severity>
          <status>open</status>
          <creator>Jack</creator>
          <created>...</created>
          <summary>Need to test Bug.add_comment()</summary>
          <comment>
            <uuid>commA</uuid>
            <short-name>/012/commA</short-name>
            <author></author>
            <date>...</date>
            <content-type>text/plain</content-type>
            <body>comment A</body>
          </comment>
          <comment>
            <uuid>commC</uuid>
            <short-name>/012/commC</short-name>
            <in-reply-to>commA</in-reply-to>
            <author></author>
            <date>...</date>
            <content-type>text/plain</content-type>
            <body>comment C</body>
          </comment>
          <comment>
            <uuid>commB</uuid>
            <short-name>/012/commB</short-name>
            <author></author>
            <date>...</date>
            <content-type>text/plain</content-type>
            <body>comment B</body>
          </comment>
        </bug>
        """
        self.add_comments([comment], **kwargs)

    def add_comments(self, comments, default_parent=None,
                     ignore_missing_references=False):
        """
        Convert a raw list of comments to single root comment.  If a
        comment does not specify a parent with .in_reply_to, the
        parent defaults to .comment_root, but you can specify another
        default parent via default_parent.
        """
        uuid_map = {}
        if default_parent == None:
            default_parent = self.comment_root
        for c in list(self.comments()) + comments:
            assert c.uuid != None
            assert c.uuid not in uuid_map
            uuid_map[c.uuid] = c
            if c.alt_id != None:
                uuid_map[c.alt_id] = c
        uuid_map[None] = self.comment_root
        uuid_map[comment.INVALID_UUID] = self.comment_root
        if default_parent != self.comment_root:
            assert default_parent.uuid in uuid_map, default_parent.uuid
        for c in comments:
            if c.in_reply_to == None \
                    and default_parent.uuid != comment.INVALID_UUID:
                c.in_reply_to = default_parent.uuid
            elif c.in_reply_to == comment.INVALID_UUID:
                c.in_reply_to = None
            try:
                parent = uuid_map[c.in_reply_to]
            except KeyError:
                if ignore_missing_references == True:
                    print >> sys.stderr, \
                        'Ignoring missing reference to %s' % c.in_reply_to
                    parent = default_parent
                    if parent.uuid != comment.INVALID_UUID:
                        c.in_reply_to = parent.uuid
                else:
                    raise comment.MissingReference(c)
            c.bug = self
            parent.append(c)

    def merge(self, other, accept_changes=True,
              accept_extra_strings=True, accept_comments=True,
              change_exception=False):
        """
        Merge info from other into this bug.  Overrides any attributes
        in self that are listed in other.explicit_attrs.

        >>> bugA = Bug(uuid='0123', summary='Need to test Bug.merge()')
        >>> bugA.date = 'Thu, 01 Jan 1970 00:00:00 +0000'
        >>> bugA.creator = 'Frank'
        >>> bugA.extra_strings += ['TAG: very helpful']
        >>> bugA.extra_strings += ['TAG: favorite']
        >>> commA = bugA.comment_root.new_reply(body='comment A')
        >>> commA.uuid = 'uuid-commA'
        >>> bugB = Bug(uuid='3210', summary='More tests for Bug.merge()')
        >>> bugB.date = 'Fri, 02 Jan 1970 00:00:00 +0000'
        >>> bugB.creator = 'John'
        >>> bugB.explicit_attrs = ['creator', 'summary']
        >>> bugB.extra_strings += ['TAG: very helpful']
        >>> bugB.extra_strings += ['TAG: useful']
        >>> commB = bugB.comment_root.new_reply(body='comment B')
        >>> commB.uuid = 'uuid-commB'
        >>> bugA.merge(bugB, accept_changes=False, accept_extra_strings=False,
        ...            accept_comments=False, change_exception=False)
        >>> print bugA.creator
        Frank
        >>> bugA.merge(bugB, accept_changes=False, accept_extra_strings=False,
        ...            accept_comments=False, change_exception=True)
        Traceback (most recent call last):
          ...
        ValueError: Merge would change creator "Frank"->"John" for bug 0123
        >>> print bugA.creator
        Frank
        >>> bugA.merge(bugB, accept_changes=True, accept_extra_strings=False,
        ...            accept_comments=False, change_exception=True)
        Traceback (most recent call last):
          ...
        ValueError: Merge would add extra string "TAG: useful" for bug 0123
        >>> print bugA.creator
        John
        >>> print bugA.extra_strings
        ['TAG: favorite', 'TAG: very helpful']
        >>> bugA.merge(bugB, accept_changes=True, accept_extra_strings=True,
        ...            accept_comments=False, change_exception=True)
        Traceback (most recent call last):
          ...
        ValueError: Merge would add comment uuid-commB (alt: None) to bug 0123
        >>> print bugA.extra_strings
        ['TAG: favorite', 'TAG: useful', 'TAG: very helpful']
        >>> bugA.merge(bugB, accept_changes=True, accept_extra_strings=True,
        ...            accept_comments=True, change_exception=True)
        >>> print bugA.xml(show_comments=True)  # doctest: +ELLIPSIS
        <bug>
          <uuid>0123</uuid>
          <short-name>/012</short-name>
          <severity>minor</severity>
          <status>open</status>
          <creator>John</creator>
          <created>...</created>
          <summary>More tests for Bug.merge()</summary>
          <extra-string>TAG: favorite</extra-string>
          <extra-string>TAG: useful</extra-string>
          <extra-string>TAG: very helpful</extra-string>
          <comment>
            <uuid>uuid-commA</uuid>
            <short-name>/012/uuid-commA</short-name>
            <author></author>
            <date>...</date>
            <content-type>text/plain</content-type>
            <body>comment A</body>
          </comment>
          <comment>
            <uuid>uuid-commB</uuid>
            <short-name>/012/uuid-commB</short-name>
            <author></author>
            <date>...</date>
            <content-type>text/plain</content-type>
            <body>comment B</body>
          </comment>
        </bug>
        """
        if hasattr(other, 'explicit_attrs'):
            for attr in other.explicit_attrs:
                old = getattr(self, attr)
                new = getattr(other, attr)
                if old != new:
                    if accept_changes:
                        setattr(self, attr, new)
                    elif change_exception:
                        raise ValueError(
                            ('Merge would change {0} "{1}"->"{2}" for bug {3}'
                             ).format(attr, old, new, self.uuid))
        for estr in other.extra_strings:
            if not estr in self.extra_strings:
                if accept_extra_strings == True:
                    self.extra_strings += [estr]
                elif change_exception == True:
                    raise ValueError, \
                        'Merge would add extra string "%s" for bug %s' \
                        % (estr, self.uuid)
        for o_comm in other.comments():
            try:
                s_comm = self.comment_root.comment_from_uuid(o_comm.uuid)
            except KeyError, e:
                try:
                    s_comm = self.comment_root.comment_from_uuid(o_comm.alt_id)
                except KeyError, e:
                    s_comm = None
            if s_comm == None:
                if accept_comments == True:
                    o_comm_copy = copy.copy(o_comm)
                    o_comm_copy.bug = self
                    o_comm_copy.id = libbe.util.id.ID(o_comm_copy, 'comment')
                    self.comment_root.add_reply(o_comm_copy)
                elif change_exception == True:
                    raise ValueError, \
                        'Merge would add comment %s (alt: %s) to bug %s' \
                        % (o_comm.uuid, o_comm.alt_id, self.uuid)
            else:
                s_comm.merge(o_comm, accept_changes=accept_changes,
                             accept_extra_strings=accept_extra_strings,
                             change_exception=change_exception)

    # methods for saving/loading/acessing settings and properties.

    def load_settings(self, settings_mapfile=None):
        if settings_mapfile == None:
            settings_mapfile = self.storage.get(
                self.id.storage('values'), '{}\n')
        try:
            settings = mapfile.parse(settings_mapfile)
        except mapfile.InvalidMapfileContents, e:
            raise Exception('Invalid settings file for bug %s\n'
                            '(BE version missmatch?)' % self.id.user())
        self._setup_saved_settings(settings)

    def save_settings(self):
        mf = mapfile.generate(self._get_saved_settings())
        self.storage.set(self.id.storage('values'), mf)

    def save(self):
        """
        Save any loaded contents to storage.  Because of lazy loading
        of comments, this is actually not too inefficient.

        However, if self.storage.is_writeable() == True, then any
        changes are automatically written to storage as soon as they
        happen, so calling this method will just waste time (unless
        something else has been messing with your stored files).
        """
        assert self.storage != None, "Can't save without storage"
        if self.bugdir != None:
            parent = self.bugdir.id.storage()
        else:
            parent = None
        self.storage.add(self.id.storage(), parent=parent, directory=True)
        self.storage.add(self.id.storage('values'), parent=self.id.storage(),
                         directory=False)
        self.save_settings()
        if len(self.comment_root) > 0:
            comment.save_comments(self)

    def load_comments(self, load_full=True):
        if load_full == True:
            # Force a complete load of the whole comment tree
            self.comment_root = self._get_comment_root(load_full=True)
        else:
            # Setup for fresh lazy-loading.  Clear _comment_root, so
            # next _get_comment_root returns a fresh version.  Turn of
            # writing temporarily so we don't write our blank comment
            # tree to disk.
            w = self.storage.writeable
            self.storage.writeable = False
            self.comment_root = None
            self.storage.writeable = w

    def remove(self):
        self.storage.recursive_remove(self.id.storage())

    # methods for managing comments

    def uuids(self):
        for comment in self.comments():
            yield comment.uuid

    def comments(self):
        for comment in self.comment_root.traverse():
            yield comment

    def new_comment(self, body=None):
        comm = self.comment_root.new_reply(body=body)
        return comm

    def comment_from_uuid(self, uuid, *args, **kwargs):
        return self.comment_root.comment_from_uuid(uuid, *args, **kwargs)

    # methods for id generation

    def sibling_uuids(self):
        if self.bugdir != None:
            return self.bugdir.uuids()
        return []


# The general rule for bug sorting is that "more important" bugs are
# less than "less important" bugs.  This way sorting a list of bugs
# will put the most important bugs first in the list.  When relative
# importance is unclear, the sorting follows some arbitrary convention
# (i.e. dictionary order).

def cmp_severity(bug_1, bug_2):
    """
    Compare the severity levels of two bugs, with more severe bugs
    comparing as less.

    >>> bugA = Bug()
    >>> bugB = Bug()
    >>> bugA.severity = bugB.severity = "wishlist"
    >>> cmp_severity(bugA, bugB) == 0
    True
    >>> bugB.severity = "minor"
    >>> cmp_severity(bugA, bugB) > 0
    True
    >>> bugA.severity = "critical"
    >>> cmp_severity(bugA, bugB) < 0
    True
    """
    if not hasattr(bug_2, "severity") :
        return 1
    return -cmp(severity_index[bug_1.severity], severity_index[bug_2.severity])

def cmp_status(bug_1, bug_2):
    """
    Compare the status levels of two bugs, with more "open" bugs
    comparing as less.

    >>> bugA = Bug()
    >>> bugB = Bug()
    >>> bugA.status = bugB.status = "open"
    >>> cmp_status(bugA, bugB) == 0
    True
    >>> bugB.status = "closed"
    >>> cmp_status(bugA, bugB) < 0
    True
    >>> bugA.status = "fixed"
    >>> cmp_status(bugA, bugB) > 0
    True
    """
    if not hasattr(bug_2, "status") :
        return 1
    val_2 = status_index[bug_2.status]
    return cmp(status_index[bug_1.status], status_index[bug_2.status])

def cmp_attr(bug_1, bug_2, attr, invert=False):
    """
    Compare a general attribute between two bugs using the
    conventional comparison rule for that attribute type.  If
    ``invert==True``, sort *against* that convention.

    >>> attr="severity"
    >>> bugA = Bug()
    >>> bugB = Bug()
    >>> bugA.severity = "critical"
    >>> bugB.severity = "wishlist"
    >>> cmp_attr(bugA, bugB, attr) < 0
    True
    >>> cmp_attr(bugA, bugB, attr, invert=True) > 0
    True
    >>> bugB.severity = "critical"
    >>> cmp_attr(bugA, bugB, attr) == 0
    True
    """
    if not hasattr(bug_2, attr) :
        return 1
    val_1 = getattr(bug_1, attr)
    val_2 = getattr(bug_2, attr)
    if val_1 == None: val_1 = None
    if val_2 == None: val_2 = None

    if invert == True :
        return -cmp(val_1, val_2)
    else :
        return cmp(val_1, val_2)

# alphabetical rankings (a < z)
cmp_uuid = lambda bug_1, bug_2 : cmp_attr(bug_1, bug_2, "uuid")
cmp_creator = lambda bug_1, bug_2 : cmp_attr(bug_1, bug_2, "creator")
cmp_assigned = lambda bug_1, bug_2 : cmp_attr(bug_1, bug_2, "assigned")
cmp_reporter = lambda bug_1, bug_2 : cmp_attr(bug_1, bug_2, "reporter")
cmp_summary = lambda bug_1, bug_2 : cmp_attr(bug_1, bug_2, "summary")
cmp_extra_strings = lambda bug_1, bug_2 : cmp_attr(bug_1, bug_2, "extra_strings")
# chronological rankings (newer < older)
cmp_time = lambda bug_1, bug_2 : cmp_attr(bug_1, bug_2, "time", invert=True)

def cmp_mine(bug_1, bug_2):
    user_id = libbe.ui.util.user.get_user_id(bug_1.storage)
    mine_1 = bug_1.assigned != user_id
    mine_2 = bug_2.assigned != user_id
    return cmp(mine_1, mine_2)

def cmp_comments(bug_1, bug_2):
    """
    Compare two bugs' comments lists.  Doesn't load any new comments,
    so you should call each bug's .load_comments() first if you want a
    full comparison.
    """
    comms_1 = sorted(bug_1.comments(), key = lambda comm : comm.uuid)
    comms_2 = sorted(bug_2.comments(), key = lambda comm : comm.uuid)
    result = cmp(len(comms_1), len(comms_2))
    if result != 0:
        return result
    for c_1,c_2 in zip(comms_1, comms_2):
        result = cmp(c_1, c_2)
        if result != 0:
            return result
    return 0

DEFAULT_CMP_FULL_CMP_LIST = \
    (cmp_status, cmp_severity, cmp_assigned, cmp_time, cmp_creator,
     cmp_reporter, cmp_comments, cmp_summary, cmp_uuid, cmp_extra_strings)

class BugCompoundComparator (object):
    def __init__(self, cmp_list=DEFAULT_CMP_FULL_CMP_LIST):
        self.cmp_list = cmp_list
    def __call__(self, bug_1, bug_2):
        for comparison in self.cmp_list :
            val = comparison(bug_1, bug_2)
            if val != 0 :
                return val
        return 0

cmp_full = BugCompoundComparator()


# define some bonus cmp_* functions
def cmp_last_modified(bug_1, bug_2):
    """
    Like cmp_time(), but use most recent comment instead of bug
    creation for the timestamp.
    """
    def last_modified(bug):
        time = bug.time
        for comment in bug.comment_root.traverse():
            if comment.time > time:
                time = comment.time
        return time
    val_1 = last_modified(bug_1)
    val_2 = last_modified(bug_2)
    return -cmp(val_1, val_2)


if libbe.TESTING == True:
    suite = doctest.DocTestSuite()



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Niall Douglas (s_sourceforge@nedprod.com) <spam@spamtrap.com>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""
Allow simple listing and loading of the various becommands and libbe
submodules (i.e. "plugins").
"""

import os
import os.path
import sys
import zipfile


_PLUGIN_PATH = os.path.realpath(
    os.path.dirname(
        os.path.dirname(
            os.path.dirname(__file__))))
if _PLUGIN_PATH not in sys.path:
    sys.path.append(_PLUGIN_PATH)

def import_by_name(modname):
    """
    >>> mod = import_by_name('libbe.bugdir')
    >>> 'BugDir' in dir(mod)
    True
    >>> import_by_name('libbe.highly_unlikely')
    Traceback (most recent call last):
      ...
    ImportError: No module named highly_unlikely
    """
    module = __import__(modname)
    components = modname.split('.')
    for comp in components[1:]:
        module = getattr(module, comp)
    return module

def zip_listdir(path, components):
    """Lists items in a directory contained in a zip file
    """
    dirpath = '/'.join(components)
    with zipfile.ZipFile(path, 'r') as f:
        return [os.path.relpath(f, dirpath) for f in f.namelist()
                if f.startswith(dirpath)]

def modnames(prefix):
    """
    >>> 'list' in [n for n in modnames('libbe.command')]
    True
    >>> 'plugin' in [n for n in modnames('libbe.util')]
    True
    """
    components = prefix.split('.')
    egg = os.path.isfile(_PLUGIN_PATH)  # are we inside a zip archive (egg)?
    if egg:
        modfiles = zip_listdir(_PLUGIN_PATH, components)
    else:
        modfiles = os.listdir(os.path.join(_PLUGIN_PATH, *components))
    # normalize .py/.pyc extensions and sort
    base_ext = [os.path.splitext(f) for f in modfiles]
    modfiles = sorted(set(
            base + '.py' for base,ext in base_ext if ext in ['.py', '.pyc']))
    for modfile in modfiles:
        if modfile.startswith('.') or not modfile:
            continue # the occasional emacs temporary file or .* file
        if modfile.endswith('.py') and modfile != '__init__.py':
            yield modfile[:-3]



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         Thomas Habets <thomas@habets.pp.se>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import sys

import libbe
import libbe.command
import libbe.command.util
import libbe.util.id
import libbe.version


class Show (libbe.command.Command):
    """Show a particular bug, comment, or combination of both.

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> io.stdout.encoding = 'ascii'
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Show(ui=ui)

    >>> ret = ui.run(cmd, args=['/a',])  # doctest: +ELLIPSIS
              ID : a
      Short name : abc/a
        Severity : minor
          Status : open
        Assigned : 
        Reporter : 
         Creator : John Doe <jdoe@example.com>
         Created : ...
    Bug A
    <BLANKLINE>

    >>> ret = ui.run(cmd, {'xml':True}, ['/a'])  # doctest: +ELLIPSIS
    <?xml version="1.0" encoding="..." ?>
    <be-xml>
      <version>
        <tag>...</tag>
        <committer>...</committer>
        <date>...</date>
        <revision>...</revision>
      </version>
      <bug>
        <uuid>a</uuid>
        <short-name>abc/a</short-name>
        <severity>minor</severity>
        <status>open</status>
        <creator>John Doe &lt;jdoe@example.com&gt;</creator>
        <created>Thu, 01 Jan 1970 00:00:00 +0000</created>
        <summary>Bug A</summary>
      </bug>
    </be-xml>
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'show'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='xml', short_name='x',
                                     help='Dump as XML'),
                libbe.command.Option(name='only-raw-body',
                    help="When printing only a single comment, just print it's"
                  " body.  This allows extraction of non-text content types."),
                libbe.command.Option(name='no-comments', short_name='c',
                    help="Disable comment output.  This is useful if you just "
                         "want more details on a bug's current status."),
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='id', metavar='ID', default=None,
                    optional=True, repeatable=True,
                    completion_callback=libbe.command.util.complete_bug_comment_id),
                ])

    def _run(self, **params):
        bugdirs = self._get_bugdirs()
        if params['only-raw-body'] == True:
            if len(params['id']) != 1:
                raise libbe.command.UserError(
                    'only one ID accepted with --only-raw-body')
            bugdir,bug,comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, params['id'][0]))
            if comment == bug.comment_root:
                raise libbe.command.UserError(
                    "--only-raw-body requires a comment ID, not '%s'"
                    % params['id'][0])
            sys.__stdout__.write(comment.body)
            return 0
        print >> self.stdout, \
            output(bugdirs, params['id'], encoding=self.stdout.encoding,
                   as_xml=params['xml'],
                   with_comments=not params['no-comments'])
        return 0

    def _long_help(self):
        return """
Show all information about the bugs or comments whose IDs are given.
If no IDs are given, show the entire repository.

Without the --xml flag set, it's probably not a good idea to mix bug
and comment IDs in a single call, but you're free to do so if you
like.  With the --xml flag set, there will never be any root comments,
so mix and match away (the bug listings for directly requested
comments will be restricted to the bug uuid and the requested
comment(s)).

Directly requested comments will be grouped by their parent bug and
placed at the end of the output, so the ordering may not match the
order of the listed IDs.
"""

def _sort_ids(bugdirs, ids, with_comments=True):
    bugs = []
    root_comments = {}
    for id in ids:
        p = libbe.util.id.parse_user(bugdirs, id)
        if p['type'] == 'bug':
            bugs.append(p['bug'])
        elif with_comments == True:
            if p['bug'] not in root_comments:
                root_comments[p['bug']] = [p['comment']]
            else:
                root_comments[p['bug']].append(p['comment'])
    for bugname in root_comments.keys():
        assert bugname not in bugs, \
            'specifically requested both #/%s/%s# and #/%s#' \
            % (bugname, root_comments[bugname][0], bugname)
    return (bugs, root_comments)

def _xml_header(encoding):
    lines = ['<?xml version="1.0" encoding="%s" ?>' % encoding,
             '<be-xml>',
             '  <version>',
             '    <tag>%s</tag>' % libbe.version.version()]
    for tag,value in sorted(libbe.version.version_info.items()):
        lines.append('    <%s>%s</%s>' % (tag, value, tag))
    lines.append('  </version>')
    return lines

def _xml_footer():
    return ['</be-xml>']

def output(bugdirs, ids, encoding, as_xml=True, with_comments=True):
    if ids == None or len(ids) == 0:
        ids = []
        for bugdir in bugdirs.values():
            bugdir.load_all_bugs()
            ids.extend([bug.id.user() for bug in bugdir])
    uuids,root_comments = _sort_ids(bugdirs, ids, with_comments)
    lines = []
    if as_xml:
        lines.extend(_xml_header(encoding))
    else:
        spaces_left = len(ids) - 1
    for bugname in uuids:
        bug = libbe.command.util.bug_from_uuid(bugdirs, bugname)
        if as_xml:
            lines.append(bug.xml(indent=2, show_comments=with_comments))
        else:
            lines.append(bug.string(show_comments=with_comments))
            if spaces_left > 0:
                spaces_left -= 1
                lines.append('') # add a blank line between bugs/comments
    for bugname,comments in root_comments.items():
        bug = libbe.command.util.bug_from_uuid(bugdirs, bugname)
        if as_xml:
            lines.extend(['  <bug>', '    <uuid>%s</uuid>' % bug.uuid])
        for commname in comments:
            try:
                comment = bug.comment_root.comment_from_uuid(commname)
            except KeyError, e:
                raise libbe.command.UserError(e.message)
            if as_xml:
                lines.append(comment.xml(indent=4))
            else:
                lines.append(comment.string())
                if spaces_left > 0:
                    spaces_left -= 1
                    lines.append('') # add a blank line between bugs/comments
        if as_xml:
            lines.append('</bug>')
    if as_xml:
        lines.extend(_xml_footer())
    return '\n'.join(lines)
